import copy
import logging
import math
import os
import random
import sys
from collections import Counter
from dataclasses import dataclass
from functools import partial
from glob import glob
from typing import List, Optional, Union

import adapters
import datasets
import numpy as np
import torch
from adapters import AdapterArguments
from adapters.models import MODEL_MIXIN_MAPPING
from adapters.models.bert.mixin_bert import BertModelAdaptersMixin
from tokenizers import AddedToken
from tqdm import tqdm
from transformers import AutoTokenizer, HfArgumentParser, TrainingArguments, set_seed

import wandb
from wtpsplit.models import SubwordXLMConfig, SubwordXLMForTokenClassification
from wtpsplit.train.adaptertrainer import AdapterTrainer
from wtpsplit.train.evaluate import evaluate_sentence
from wtpsplit.train.train import collate_fn, setup_logging
from wtpsplit.train.trainer import Trainer
from wtpsplit.train.utils import Model
from wtpsplit.utils import Constants, LabelArgs, get_label_dict, get_subword_label_dict

logger = logging.getLogger(__name__)

os.environ["TOKENIZERS_PARALLELISM"] = "false"

MODEL_MIXIN_MAPPING["SubwordXLMRobertaModel"] = BertModelAdaptersMixin


@dataclass
class Args:
    model_name_or_path: str
    base_model: str = "xlm-roberta-base"
    shuffle: bool = True
    text_path: str = "data/all_data_11_05-lyrics.pth"
    include_languages: List[str] = None
    preprocessing_num_workers: int = 1
    block_size: int = 512
    overflow_size: int = 16
    eval_stride: int = 256
    loss_margin: float = 0.5
    pack_samples: bool = False
    one_sample_per_line: bool = False
    use_loss_weights: bool = False
    do_sentence_training: bool = True
    do_auxiliary_training: bool = False
    aux_training_weight: float = 1.0
    ignore_non_hyphen: bool = False
    non_punctuation_sample_ratio: float = None
    adapter_warmup_steps: int = 0
    adapter_lr_multiplier: float = 1.0
    text_column: str = "text"
    num_hidden_layers: int = 0

    # NEW PARAMS
    use_subwords: bool = False
    freeze_classifier: bool = False
    clf_from_scratch: bool = False
    unfreeze_ln: bool = False
    do_process: bool = False
    meta_clf: bool = False
    wandb_project: str = "sentence"
    eval_every: int = 5
    # corruption
    skip_eval_loss: bool = False
    subsample: Optional[float] = None


def main():
    parser = HfArgumentParser([Args, TrainingArguments, LabelArgs, AdapterArguments])
    if sys.argv[1].endswith(".json"):
        (args, training_args, label_args, adapter_args) = parser.parse_json_file(sys.argv[1])
        wandb_name = training_args.output_dir
    else:
        (args, training_args, label_args, adapter_args) = parser.parse_args_into_dataclasses()
        wandb_name = None

    setup_logging(training_args)
    set_seed(training_args.seed)

    num_labels = Constants.AUX_OFFSET + (
        (1 + len(Constants.PUNCTUATION_CHARS))
        if (label_args.use_auxiliary or args.do_auxiliary_training or args.meta_clf)
        else 0
    )

    if args.num_hidden_layers:
        config = SubwordXLMConfig.from_pretrained(
            args.model_name_or_path,
            num_labels=num_labels,
            num_hidden_layers=args.num_hidden_layers,
        )
    else:
        config = SubwordXLMConfig.from_pretrained(args.model_name_or_path, num_labels=num_labels)

    def prepare_dataset(
        data,
        num_workers=1,
        include_languages=None,
        dataset_name="ud",
        shuffle=False,
        split="train",
        subsample: Union[None, int, float] = None,
        one_sample_per_line: bool = False,
    ):
        with training_args.main_process_first():
            # this only uses 1 dataset-language combination, but can be easily adapted if needed
            for lang in include_languages:
                if split == "train":
                    dataset = data[lang]["sentence"][dataset_name]["meta"]["train_data"]
                elif split == "valid":
                    dataset = data[lang]["sentence"][dataset_name]["data"]
                if dataset is None:
                    return None

                if one_sample_per_line or isinstance(dataset[0], list):
                    processed_dataset = []
                    for chunk in dataset:
                        processed_chunk = {}
                        processed_chunk["lang"] = lang
                        processed_chunk["ends_with_punctuation"] = chunk[-1].endswith(
                            tuple(Constants.PUNCTUATION_CHARS)
                        )
                        # join all chunks
                        processed_chunk[args.text_column] = "\n".join(chunk)
                        processed_dataset.append(processed_chunk)
                    dataset = datasets.Dataset.from_list(processed_dataset)

                else:
                    dataset = datasets.Dataset.from_list(
                        [
                            {
                                args.text_column: sample + "\n" if sample and sample[-1] != "\n" else sample,  # TODO
                                "lang": lang,
                                "ends_with_punctuation": sample.endswith(tuple(Constants.PUNCTUATION_CHARS)),
                            }
                            for sample in dataset
                        ]
                    )
            with training_args.main_process_first():
                logger.warning(f"Loaded {len(dataset)} examples for {lang} {dataset_name} {split} dataset.")

        if shuffle:
            dataset = dataset.shuffle(seed=training_args.seed)
        if subsample:
            old_length = len(dataset)
            if isinstance(subsample, int):
                # ensure that we don't try to select more than the dataset length
                subsample = min(subsample, len(dataset))
                dataset = dataset.select(range(subsample))
            elif isinstance(subsample, float):
                dataset = dataset.select(range(int(subsample * len(dataset))))
            logger.warning(f"Subsampled {len(dataset)} examples from {old_length}.")

        # ignore
        if args.ignore_non_hyphen:
            with training_args.main_process_first():
                dataset = dataset.filter(
                    lambda sample: any(c in sample[args.text_column] for c in label_args.hyphen_chars),
                    num_proc=args.preprocessing_num_workers,
                )
                with training_args.main_process_first():
                    logger.info(f"Filtered to {len(dataset)} examples.")

        # "punctuation-specific sampling" in the WtP paper
        if args.non_punctuation_sample_ratio is not None:
            languages_without_punctuation = {
                lang_code
                for lang_code in Constants.LANGINFO.index
                if Constants.LANGINFO.loc[lang_code, "no_punctuation"]
            }

            def drop_some_non_punctuation_samples(examples):
                include_indices = set(
                    np.where([lang_code not in languages_without_punctuation for lang_code in examples["lang"]])[0]
                )
                punctuation_indices = {
                    i for i in np.where(examples["ends_with_punctuation"])[0] if i in include_indices
                }

                target_n_non_punct = int(
                    (len(punctuation_indices) * args.non_punctuation_sample_ratio)
                    / (1 - args.non_punctuation_sample_ratio)
                )
                n_drop = (len(include_indices) - len(punctuation_indices)) - target_n_non_punct

                out = [True for _ in range(len(examples["ends_with_punctuation"]))]

                if n_drop <= 0:
                    return out
                drop_indices = np.random.choice(
                    list(include_indices - punctuation_indices),
                    n_drop,
                    replace=False,
                )

                for i in drop_indices:
                    out[i] = False

                return out

            with training_args.main_process_first():
                dataset = dataset.filter(
                    drop_some_non_punctuation_samples,
                    batched=True,
                    batch_size=1_000_000,
                    num_proc=num_workers,
                )

        def tokenize_texts(examples):
            # do not return CLS and SEP token here
            # there should only be 1 of these per block later, not multiple
            # we still can't use return_special_tokens=False since we need the \n token later for the labels
            tokenized = tokenizer(examples[args.text_column], verbose=False)
            return {"input_ids": [example[1:-1] for example in tokenized["input_ids"]]}

        # similar to group_texts in huggingface's run_clm.py / run_mlm.py: https://github.com/huggingface/transformers/blob/main/examples/pytorch/language-modeling/run_mlm.py
        def group_texts(examples):
            all_input_blocks = []
            all_input_block_lengths = []
            all_langs = []

            def maybe_pad(text):
                if args.pack_samples:
                    padding = model.backbone.config.downsampling_rate - (len(text) % model.backbone.downsampling_rate)
                    if padding == model.backbone.downsampling_rate:
                        padding = 0

                    text += chr(0) * padding

                return text

            for current_lang in set(examples["lang"]):
                if not args.use_subwords:
                    lang_texts = [
                        maybe_pad(text)
                        for text, lang in zip(examples["input_ids"], examples["lang"])
                        if lang == current_lang
                    ]
                else:
                    # only retain current_lang examples (all columns)
                    lang_subwords = [
                        subwords
                        for subwords, lang in zip(examples["input_ids"], examples["lang"])
                        if lang == current_lang
                    ]
                    # filter out some special tokens
                    # from html tags, mostly in Latin, Thai & Korean
                    lang_subwords = [
                        [subword for subword in subwords if subword not in special_tokens_ids]
                        for subwords in lang_subwords
                    ]
                # pack_samples used for the compound part, so irrelevant
                if args.pack_samples:
                    if args.use_subwords:
                        raise NotImplementedError
                    blocks = []
                    block_ids = []

                    current_block = ["", []]

                    for i, text in enumerate(lang_texts):
                        if len(text) > args.block_size:
                            continue

                        current_block[0] += text
                        current_block[1] += [i] * len(text)

                        if i + 1 < len(lang_texts) and len(current_block[0]) + len(lang_texts[i + 1]) > args.block_size:
                            padding = args.block_size - len(current_block[0])

                            current_block[0] += chr(0) * padding
                            current_block[1] += [i] * padding
                            blocks.append(current_block[0])
                            block_ids.append(current_block[1])

                            current_block = ["", []]

                    if len(current_block[0]) > 0:
                        padding = args.block_size - len(current_block[0])

                        current_block[0] += chr(0) * padding
                        current_block[1] += [i] * padding
                        blocks.append(current_block[0])
                        block_ids.append(current_block[1])
                else:
                    if not args.use_subwords:
                        concatenated_texts = "".join(lang_texts)
                        concatenated_ids = [i for i, text in enumerate(lang_texts) for _ in text]
                    else:
                        # concatenate token lists
                        concatenated_texts = [item for sublist in lang_subwords for item in sublist]
                        concatenated_ids = [i for i, subwords in enumerate(lang_subwords) for _ in subwords]

                    total_length = len(concatenated_texts)

                    best_length = math.ceil(total_length / args.block_size) * args.block_size + args.overflow_size
                    while best_length > total_length:
                        best_length -= args.block_size
                    if best_length < args.block_size:
                        best_length = args.block_size + 1

                    if best_length < 0:
                        continue

                    concatenated_texts = concatenated_texts[:best_length]
                    concatenated_ids = concatenated_ids[:best_length]

                    blocks = [
                        concatenated_texts[i : i + args.block_size + args.overflow_size]
                        for i in range(0, best_length - args.block_size, args.block_size)
                    ]
                    block_ids = [
                        concatenated_ids[i : i + args.block_size + args.overflow_size]
                        for i in range(0, best_length - args.block_size, args.block_size)
                    ]

                block_langs = [current_lang] * len(blocks)

                all_input_blocks.extend(blocks)
                all_input_block_lengths.extend([list(Counter(ids).values()) for ids in block_ids])
                all_langs.extend(block_langs)

            return {
                "input_ids": all_input_blocks,
                "block_lengths": all_input_block_lengths,
                "lang": all_langs,
            }

        if args.pack_samples:
            assert not one_sample_per_line

        if args.use_subwords:
            with training_args.main_process_first():
                dataset = dataset.map(
                    tokenize_texts,
                    batched=True,
                    num_proc=num_workers,
                    remove_columns=[args.text_column],
                    desc="Tokenizing",
                )
        else:
            # this is no longer used and would cause an error otherwise
            with training_args.main_process_first():
                dataset = dataset.rename_column(args.text_column, "input_ids")

        if not one_sample_per_line:
            with training_args.main_process_first():
                dataset = dataset.map(
                    group_texts,
                    batched=True,
                    num_proc=num_workers,
                    # a bit hacky but oh well, only drop if sentence
                    remove_columns=["ends_with_punctuation"] if args.text_column == "text" else [],
                    desc="Grouping",
                )

        return dataset

    with training_args.main_process_first():
        data = torch.load(
            args.text_path,
        )
        # sort alphabetically by key for alphabetical filtering
        data = dict(sorted(data.items()))

    if not args.include_languages:
        args.include_languages = list(data.keys())  # use all

    # 1 wandb run for all language-dataset combinations
    if "wandb" in training_args.report_to and training_args.process_index == 0:
        wandb.init(name=wandb_name, project=args.wandb_project, group=wandb_name)
        wandb.config.update(args)
        wandb.config.update(training_args)
        wandb.config.update(label_args)
        wandb.config.update(adapter_args)

        for file in glob(os.path.join(os.path.dirname(__file__), "*.py")):
            wandb.save(os.path.abspath(file), policy="now")

    for lang in tqdm(data.keys(), desc="Language"):
        if lang in args.include_languages:
            for dataset_name in data[lang]["sentence"].keys():
                print("RUNNING:", dataset_name, lang)
                # do model stuff here; otherwise, head params would be overwritten every time
                backbone = SubwordXLMForTokenClassification.from_pretrained(
                    args.model_name_or_path, config=copy.deepcopy(config), ignore_mismatched_sizes=True
                )
                backbone.config.base_model = args.base_model

                # setup adapters
                model_type = backbone.config.model_type
                # adapters need xlm-roberta as model type.
                backbone.config.model_type = "xlm-roberta"  # needed for adapter setup
                # adapters.init(backbone)
                # reset model type (used later)
                backbone.config.model_type = model_type

                tokenizer = AutoTokenizer.from_pretrained(args.base_model)
                # needed since we create labels in collate_fn based on tokens
                tokenizer.add_special_tokens({"additional_special_tokens": [AddedToken("\n")]})
                custom_token_id = tokenizer.convert_tokens_to_ids("\n")
                # used later to filter out special tokens
                special_tokens_ids = set(tokenizer.all_special_ids)
                special_tokens_ids.discard(custom_token_id)

                if "short" in dataset_name:
                    one_sample_per_line = True
                else:
                    one_sample_per_line = args.one_sample_per_line  # used for lyrics

                model = Model(
                    backbone,
                    loss_margin=args.loss_margin,
                    use_loss_weights=args.use_loss_weights,
                    do_sentence_training=args.do_sentence_training,
                    do_auxiliary_training=args.do_auxiliary_training,
                    aux_training_weight=args.aux_training_weight,
                )

                with training_args.main_process_first():
                    valid_dataset = prepare_dataset(
                        data=data,
                        num_workers=1,
                        include_languages=[lang],
                        dataset_name=dataset_name,
                        shuffle=False,
                        split="valid",
                        one_sample_per_line=one_sample_per_line,
                    )
                    logger.warning(f"Valid ds for {lang} {dataset_name} has {len(valid_dataset)} examples.")

                    train_dataset = prepare_dataset(
                        data=data,
                        num_workers=args.preprocessing_num_workers,
                        include_languages=[lang],
                        dataset_name=dataset_name,
                        shuffle=args.shuffle,
                        split="train",
                        subsample=args.subsample,
                        one_sample_per_line=one_sample_per_line,
                    )
                    if train_dataset is None or valid_dataset is None:
                        logger.warning(f"Skipping {lang} {dataset_name} due to missing data.")
                        continue
                    logger.warning(f"Train ds for {lang} {dataset_name} has {len(train_dataset)} examples.")

                # print some samples from the dataset
                count = 0
                while count < 1:
                    index = random.choice(range(len(train_dataset)))
                    sample = train_dataset[index]

                    logger.warning(f"Sample {index} of the training set: {sample}.")
                    if tokenizer:
                        logger.warning(tokenizer.decode(sample["input_ids"]))
                    count += 1

                def compute_metrics(trainer):
                    metrics = {}
                    eval_data = data[lang]["sentence"][dataset_name]["data"]

                    model = trainer._wrap_model(trainer.model, training=False)

                    with training_args.main_process_first():
                        # feeding in single samples is too slow --> feed in as one long text
                        if args.one_sample_per_line:
                            eval_data = [item for sublist in eval_data for item in sublist]
                        elif isinstance(eval_data[0], list):
                            eval_data = [item for sublist in eval_data for item in sublist]
                        score, info = evaluate_sentence(
                            lang,
                            eval_data,
                            model,
                            stride=args.eval_stride,
                            block_size=args.block_size,
                            batch_size=training_args.per_device_eval_batch_size,
                        )
                        metrics[f"{dataset_name}/{lang}/pr_auc"] = score
                        metrics[f"{dataset_name}/{lang}/f1"] = info["f1"]
                        metrics[f"{dataset_name}/{lang}/f1_best"] = info["f1_best"]
                        metrics[f"{dataset_name}/{lang}/threshold_best"] = info["threshold_best"]

                    return metrics

                label_dict = (
                    get_subword_label_dict(label_args, tokenizer) if args.use_subwords else get_label_dict(label_args)
                )

                # if adapter_args.train_adapter:
                #     # init new adapter
                #     # model.backbone.add_adapter(
                #     #     "text", config=adapter_args.adapter_config, set_active=True, overwrite_ok=True
                #     # )
                #     model.backbone.train_adapter("text")
                #     kwargs = {"logging_prefix": f"{dataset_name}/{lang}/", "skip_eval_loss": args.skip_eval_loss}
                # else:
                    # # needed in the trainer otherwise (WtP-related only)
                    # training_args.adapter_warmup_steps = args.adapter_warmup_steps
                    # training_args.adapter_lr_multiplier = args.adapter_lr_multiplier
                kwargs = {}

                # with training_args.main_process_first():
                #     logger.warning(model.backbone.adapter_summary())

                if args.freeze_classifier:
                    for n, p in model.backbone.named_parameters():
                        if "classifier" in n:
                            p.requires_grad = False
                # not done: keeping clf head is much better
                if args.clf_from_scratch:
                    model.backbone.classifier = torch.nn.Linear(model.backbone.config.hidden_size, num_labels)

                # eval only 5x during the entire training
                training_args.evaluation_strategy = "steps"
                training_args.eval_steps = max(
                    (
                        len(train_dataset)
                        // training_args.per_device_train_batch_size
                        // training_args.gradient_accumulation_steps
                        // args.eval_every
                        * training_args.num_train_epochs
                    ),
                    args.eval_every,
                )

                trainer_cls = Trainer

                trainer = trainer_cls(
                    model,
                    training_args,
                    train_dataset=train_dataset,
                    eval_dataset=valid_dataset,
                    compute_metrics=compute_metrics,
                    data_collator=partial(
                        collate_fn,
                        args=args,
                        label_args=label_args,
                        label_dict=label_dict,
                        tokenizer=tokenizer,
                        add_lang_ids=False,
                    ),
                    **kwargs,
                )
                trainer.train(resume_from_checkpoint=training_args.resume_from_checkpoint)
                logger.warning(f"Finished training for {lang} {dataset_name}.")

                # only save trained module
                if training_args.local_rank == 0:
                    if not os.path.exists(os.path.join(training_args.output_dir, dataset_name, lang)):
                        os.makedirs(os.path.join(training_args.output_dir, dataset_name, lang))
                    save_model = copy.deepcopy(model.backbone)
                    save_model = save_model.to("cpu")
                    if adapter_args.train_adapter:
                        save_model.to("cpu").save_adapter(
                            adapter_name="text",
                            save_directory=os.path.join(training_args.output_dir, dataset_name, lang),
                            with_head=True,
                        )
                    else:
                        save_model.to("cpu").save_pretrained(os.path.join(training_args.output_dir, dataset_name, lang))


def _mp_fn(index):
    # For xla_spawn (TPUs)
    main()


if __name__ == "__main__":
    main()



import subprocess
import unicodedata
import os

import numpy as np
import regex as re
import torch
from sklearn import linear_model
from sklearn.metrics import f1_score, precision_recall_curve, precision_score, recall_score

from wtpsplit.extract import extract
from wtpsplit.utils import Constants, indices_to_sentences, lang_code_to_lang, reconstruct_sentences


def preprocess_sentence(sentence):
    # right-to-left-mark
    sentence = sentence.replace(chr(8207), "")

    return re.sub(r"\s+", " ", unicodedata.normalize("NFKC", sentence.lstrip("-").strip()))


def get_labels(lang_code, sentences, after_space=True):
    separator = Constants.SEPARATORS[lang_code]
    text = separator.join(sentences)

    true_end_indices = np.cumsum(np.array([len(s) for s in sentences])) + np.arange(1, len(sentences) + 1) * len(
        separator
    )

    if not after_space:
        true_end_indices -= len(separator) + 1
    else:
        # no space after last
        true_end_indices[-1] -= len(separator)

    labels = np.zeros(len(text) + 1)
    labels[true_end_indices] = 1

    return labels


def evaluate_sentences(
    lang_code, sentences, predicted_sentences, return_indices: bool = False, exclude_every_k: int = 0
):
    separator = Constants.SEPARATORS[lang_code]

    text = separator.join(sentences)

    assert len(text) == len("".join(predicted_sentences))

    labels = get_labels(lang_code, sentences)

    predicted_end_indices = np.cumsum(np.array([len(s) for s in predicted_sentences]))
    predictions = np.zeros_like(labels)
    predictions[predicted_end_indices] = 1

    assert len(labels) == len(predictions)

    # we exclude labels for metrics calculation to enable a fair comparison with LLMs.
    if exclude_every_k > 0:
        true_end_indices = np.where(labels == 1)[0]
        # every k-th from those where labels are 1
        indices_to_remove = true_end_indices[exclude_every_k - 1 :: exclude_every_k]

        # mask for indices to keep
        mask = np.ones_like(labels, dtype=bool)
        mask[indices_to_remove] = False
        mask[-1] = False  # last is always excluded

        # remove indices
        labels = labels[mask]
        predictions = predictions[mask]

        assert len(labels) == len(predictions)

    return f1_score(labels, predictions, zero_division=0), {
        "recall": recall_score(labels, predictions, zero_division=0),
        "precision": precision_score(labels, predictions, zero_division=0),
        # pairwise: ignore end-of-text label
        # only correct if we correctly predict the single newline in between the sentence pair
        # --> no false positives, no false negatives allowed!
        "correct_pairwise": np.all(labels[:-1] == predictions[:-1]),
        "true_indices": np.where(labels)[0].tolist() if return_indices else None,
        "predicted_indices": np.where(predictions)[0].tolist() if return_indices else None,
        "length": len(labels),
    }


def evaluate_sentences_llm(labels, predictions, return_indices: bool = False, exclude_every_k: int = 0):
    assert len(labels) == len(predictions)

    if exclude_every_k > 0:
        true_end_indices = np.where(labels == 1)[0]
        # every k-th from those where labels are 1
        indices_to_remove = true_end_indices[exclude_every_k - 1 :: exclude_every_k]

        # mask for indices to keep
        mask = np.ones_like(labels, dtype=bool)
        mask[indices_to_remove] = False
        mask[-1] = False  # last is always excluded

        # remove indices
        labels = labels[mask]
        predictions = predictions[mask]

        assert len(labels) == len(predictions)

    return {
        "f1": f1_score(labels, predictions, zero_division=0),
        "recall": recall_score(labels, predictions, zero_division=0),
        "precision": precision_score(labels, predictions, zero_division=0),
        # pairwise: ignore end-of-text label
        # only correct if we correctly predict the single newline in between the sentence pair
        # --> no false positives, no false negatives allowed!
        "correct_pairwise": int(np.all(labels[:-1] == predictions[:-1])),
        "true_indices": np.where(labels)[0].tolist() if return_indices else None,
        "predicted_indices": np.where(predictions)[0].tolist() if return_indices else None,
        "length": len(labels),
    }


def train_mixture(lang_code, original_train_x, train_y, n_subsample=None, features=None, skip_punct: bool = False):
    original_train_x = torch.from_numpy(original_train_x).float()

    train_y = train_y[:-1]

    if original_train_x.shape[1] > Constants.AUX_OFFSET and not skip_punct:
        if features is not None:
            train_x = original_train_x[:, features]
        else:
            train_x = original_train_x

        train_x = train_x.float()

        clf = linear_model.LogisticRegression(max_iter=10_000, random_state=0)
        clf.fit(train_x, train_y)
        preds = clf.predict_proba(train_x)[:, 1]

        p, r, t = precision_recall_curve(train_y, preds)
        f1 = 2 * p * r / (p + r + 1e-6)
        best_threshold_transformed = t[f1.argmax()]
    else:
        clf = None
        best_threshold_transformed = None

    p, r, t = precision_recall_curve(train_y, torch.sigmoid(original_train_x[:, Constants.NEWLINE_INDEX]))
    f1 = 2 * p * r / (p + r + 1e-6)
    best_threshold_newline = t[f1.argmax()]
    print(best_threshold_transformed, best_threshold_newline)

    return clf, features, best_threshold_transformed, best_threshold_newline


def evaluate_mixture(
    lang_code,
    test_x,
    true_sentences,
    return_indices,
    exclude_every_k,
    clf,
    features,
    threshold_transformed,
    threshold_newline,
):
    test_x = torch.from_numpy(test_x)
    text = Constants.SEPARATORS[lang_code].join(true_sentences)

    predicted_indices_newline = np.where(
        torch.sigmoid(test_x[..., Constants.NEWLINE_INDEX].float()).numpy() > threshold_newline
    )[0]

    if clf is not None:
        if features is not None:
            test_x = test_x[:, features]

        test_x = test_x.float()

        probs = clf.predict_proba(test_x)[:, 1]
        predicted_indices_transformed = np.where(probs > threshold_transformed)[0]
    else:
        predicted_indices_transformed = None

    score_newline, info_newline = evaluate_sentences(
        lang_code,
        true_sentences,
        reconstruct_sentences(text, indices_to_sentences(text, predicted_indices_newline)),
        return_indices,
        exclude_every_k,
    )

    indices_newline_info = {
        "true_indices": info_newline.pop("true_indices"),
        "pred_indices": info_newline.pop("predicted_indices"),
        "length": info_newline.pop("length"),
    }

    if predicted_indices_transformed is None:
        return (
            score_newline,
            None,
            {"info_newline": info_newline, "info_transformed": None},
            indices_newline_info,
            None,
        )

    score_transformed, info_transformed = evaluate_sentences(
        lang_code,
        true_sentences,
        reconstruct_sentences(text, indices_to_sentences(text, predicted_indices_transformed)),
        return_indices,
        exclude_every_k,
    )
    indices_transformed_info = {
        "true_indices": info_transformed.pop("true_indices"),
        "pred_indices": info_transformed.pop("predicted_indices"),
        "length": info_transformed.pop("length"),
    }

    return (
        score_newline,
        score_transformed,
        {
            "info_newline": info_newline,
            "info_transformed": info_transformed,
        },
        indices_newline_info,
        indices_transformed_info,
    )


def our_sentencize(
    sentence_model,
    lang_code,
    text,
    clf,
    features,
    threshold_transformed,
    threshold_newline,
    newline_only=False,
    block_size=512,
    stride=64,
    batch_size=32,
):
    logits = extract(
        [text],
        sentence_model,
        lang_code=lang_code,
        stride=stride,
        max_block_size=block_size,
        batch_size=batch_size,
        pad_last_batch=False,
        use_hidden_states=False,
        verbose=False,
    )[0]

    predicted_indices_newline = np.where(
        torch.sigmoid(logits[..., Constants.NEWLINE_INDEX].float()).numpy() > threshold_newline
    )[0]

    if features is not None:
        x = logits[:, features]
    else:
        x = logits

    x = x.float()

    probs = clf.predict_proba(x)[:, 1]
    predicted_indices_transformed = np.where(probs > threshold_transformed)[0]

    if newline_only:
        return reconstruct_sentences(text, indices_to_sentences(text, predicted_indices_newline))
    else:
        return reconstruct_sentences(text, indices_to_sentences(text, predicted_indices_transformed))


###########
# BASELINES
###########

ERSATZ_LANGUAGES = {
    "ar",
    "cs",
    "de",
    "en",
    "es",
    "et",
    "fi",
    "fr",
    "gu",
    "hi",
    "iu",
    "ja",
    "kk",
    "km",
    "lt",
    "lv",
    "pl",
    "ps",
    "ro",
    "ru",
    "ta",
    "tr",
    "zh",
}


class LanguageError(ValueError):
    pass


def ersatz_sentencize(
    lang_code,
    text,
    infile="notebooks/data/tmp/in.tmp",
    outfile="notebooks/data/tmp/out.tmp",
):
    if lang_code not in ERSATZ_LANGUAGES:
        raise LanguageError(f"ersatz does not support {lang_code}")

    # check if infile parent dir exists, if not, create it
    if not os.path.exists(os.path.dirname(infile)):
        os.makedirs(os.path.dirname(infile))
    # check if outfile parent dir exists, if not, create it
    if not os.path.exists(os.path.dirname(outfile)):
        os.makedirs(os.path.dirname(outfile))

    open(infile, "w").write(text)

    subprocess.check_output(
        f"cat {infile} | ersatz --quiet -m default-multilingual > {outfile}",
        shell=True,
    )

    return reconstruct_sentences(text, open(outfile).readlines())


def pysbd_sentencize(lang_code, text):
    import pysbd

    try:
        return reconstruct_sentences(text, pysbd.Segmenter(language=lang_code, clean=False).segment(text))
    except ValueError:
        raise LanguageError(f"pysbd does not support {lang_code}")


SPACY_LANG_TO_DP_MODEL = {
    "ca": "ca_core_news_sm",
    "zh": "zh_core_web_sm",
    "hr": "hr_core_news_sm",
    "da": "da_core_news_sm",
    "nl": "nl_core_news_sm",
    "en": "en_core_web_sm",
    "fi": "fi_core_news_sm",
    "fr": "fr_core_news_sm",
    "de": "de_core_news_sm",
    "el": "el_core_news_sm",
    "it": "it_core_news_sm",
    "ja": "ja_core_news_sm",
    "ko": "ko_core_news_sm",
    "lt": "lt_core_news_sm",
    "mk": "mk_core_news_sm",
    "nb": "nb_core_news_sm",
    "pl": "pl_core_news_sm",
    "pt": "pt_core_news_sm",
    "ro": "ro_core_news_sm",
    "ru": "ru_core_news_sm",
    "es": "es_core_news_sm",
    "sv": "sv_core_news_sm",
    "uk": "uk_core_news_sm",
    "xx": "xx_sent_ud_sm",
}


def spacy_sent_sentencize(lang_code, text):
    import spacy

    try:
        nlp = spacy.blank(lang_code)
        nlp.add_pipe("sentencizer")
        nlp.max_length = 1_000_000_000

        if lang_code == "ja":
            # spacy uses SudachiPy for japanese, which has a length limit:
            # https://github.com/WorksApplications/sudachi.rs/blob/c7d20b22c68bb3f6585351847ae91bc9c7a61ec5/sudachi/src/input_text/buffer/mod.rs#L124-L126
            # so we need to chunk the input and sentencize the chunks separately
            chunksize = 10_000
            chunks = []
            for i in range(0, len(text), chunksize):
                chunks.append(text[i : i + chunksize])

            assert sum(len(c) for c in chunks) == len(text)
            return reconstruct_sentences(text, [str(s) for c in chunks for s in nlp(c).sents])

        return reconstruct_sentences(text, list([str(s) for s in nlp(text).sents]))
    except ImportError:
        raise LanguageError(f"spacy_sent does not support {lang_code}")


def spacy_dp_sentencize(lang_code, text):
    import spacy

    try:
        nlp = spacy.load(SPACY_LANG_TO_DP_MODEL[lang_code], disable=["ner"])
        nlp.max_length = 1_000_000_000

        if lang_code == "ja":
            # spacy uses SudachiPy for japanese, which has a length limit:
            # https://github.com/WorksApplications/sudachi.rs/blob/c7d20b22c68bb3f6585351847ae91bc9c7a61ec5/sudachi/src/input_text/buffer/mod.rs#L124-L126
            # so we need to chunk the input and sentencize the chunks separately
            chunksize = 10_000
            chunks = []
            for i in range(0, len(text), chunksize):
                chunks.append(text[i : i + chunksize])

            assert sum(len(c) for c in chunks) == len(text)
            return reconstruct_sentences(text, [str(s) for c in chunks for s in nlp(c).sents])

        return reconstruct_sentences(text, list([str(s) for s in nlp(text).sents]))
    except KeyError:
        raise LanguageError(f"spacy_dp does not support {lang_code}")


def punkt_sentencize(lang_code, text):
    from nltk.tokenize import sent_tokenize

    try:
        return reconstruct_sentences(text, sent_tokenize(text, language=lang_code_to_lang(lang_code).lower()))
    except LookupError:
        raise LanguageError(f"punkt does not support {lang_code}")



import math
import sys
import logging

import numpy as np
from transformers import AutoTokenizer
from tokenizers import AddedToken

from wtpsplit.utils import Constants, hash_encode
from wtpsplit.extract import BertCharORTWrapper

logger = logging.getLogger(__name__)


def extract_batched(
    batch_of_texts,
    model,
    block_size,
    batch_size,
    lang_code=None,
    pad_last_batch=False,
    verbose=False,
):
    """
    Like extract.py, but does not split the input into chunks of block_size.
    Instead, it processes the input in batches. So each input batch must be smaller than block_size.
    """
    if "xlm" in model.config.model_type:
        use_subwords = True
        tokenizer = AutoTokenizer.from_pretrained("xlm-roberta-base")
        tokenizer.add_special_tokens({"additional_special_tokens": [AddedToken("\n")]})
        tokens = tokenizer(
            batch_of_texts,
            return_offsets_mapping=True,
            verbose=False,
            padding="max_length",  # pad to max length (TPUs need fixed length inputs)
            return_tensors="np",
            truncation=True,
            max_length=block_size,
        )
        attention_mask = tokens["attention_mask"]
        input_ids = tokens["input_ids"]
        offset_mapping = [offset[1:-1] for offset in tokens.pop("offset_mapping")]
        pad_token_id = tokenizer.pad_token_id
        text_lengths = [len(text) for text in input_ids]
    else:
        pad_token_id = 0
        use_subwords = False
        input_ids = batch_of_texts.copy()
        text_lengths = [len(text) for text in input_ids]
        offset_mapping = text_lengths

    # when using ChLMs, it can be that the input is too long --> simply truncate
    if max(text_lengths) > block_size:
        # truncate
        longer_than_block_size = [i for i, length in enumerate(text_lengths) if length > block_size]
        if verbose:
            logger.info(f"Truncating {len(longer_than_block_size)} texts longer than block_size={block_size}")
        for i in longer_than_block_size:
            input_ids[i] = input_ids[i][:block_size]

    # make sure block_size is a multiple of downsampling rate
    downsampling_rate = getattr(model.config, "downsampling_rate", 1)
    block_size = math.ceil(block_size / downsampling_rate) * downsampling_rate

    if not use_subwords:
        codec = "utf-32-le" if sys.byteorder == "little" else "utf-32-be"
        hashed_ids = []
        attention_mask = np.ones((len(input_ids), block_size), dtype=int)
        for i, text in enumerate(input_ids):
            ord = np.frombuffer(bytearray(text, encoding=codec), dtype=np.int32)
            # pad
            if len(ord) < block_size:
                # mask out padding
                attention_mask[i, len(ord) :] = 0
                # pad to max length, i.e., block size (no truncation due to TPUs)
                ord = np.pad(ord, (0, block_size - len(ord)))
            hashed_ids.append(hash_encode(ord, model.config.num_hash_functions, model.config.num_hash_buckets))
        hashed_ids = np.array(hashed_ids)

    uses_lang_adapters = getattr(model.config, "language_adapter", "off") == "on"
    if uses_lang_adapters:
        if lang_code is None:
            raise ValueError("Please specify a `lang_code` when using a model with language adapters.")

        if isinstance(model, BertCharORTWrapper):
            raise ValueError("Language adapters are not supported in ONNX models.")

        language_ids = np.array(
            [Constants.LANG_CODE_TO_INDEX[lang_code]] * batch_size,
            dtype=int,
        )
    else:
        language_ids = None

    if len(attention_mask) < batch_size and pad_last_batch:
        n_missing = batch_size - len(attention_mask)

        if not use_subwords:
            hashed_ids = np.pad(hashed_ids, ((0, n_missing), (0, 0), (0, 0)))
        else:
            # Pad with the specific pad_token_id for the tokenizer
            input_ids = np.pad(input_ids, ((0, n_missing), (0, 0)), constant_values=pad_token_id)
        attention_mask = np.pad(attention_mask, ((0, n_missing), (0, 0)))
    else:
        n_missing = 0

    kwargs = {"language_ids": language_ids} if uses_lang_adapters else {}

    if use_subwords and model.config.model_type == "xlm-roberta":
        # TODO: generalize
        import torch

        with torch.no_grad():
            logits = (
                model.model(
                    input_ids=torch.from_numpy(input_ids).to(model.model.device),
                    attention_mask=torch.from_numpy(attention_mask).to(model.model.device),
                    **kwargs,
                )["logits"]
                .cpu()
                .numpy()
            )
    else:
        logits = model(
            input_ids=input_ids if use_subwords else None,
            hashed_ids=None if use_subwords else hashed_ids,
            attention_mask=attention_mask,
            **kwargs,
        )["logits"]
    if use_subwords:
        logits = logits[:, 1:-1, :]  # remove CLS and SEP tokens

    return (
        logits[: batch_size - n_missing],
        offset_mapping,
        tokenizer if use_subwords else None,
    )



import copy
import json
import logging
import os
import sys
import time
from dataclasses import dataclass
from typing import List, Union

import h5py
import numpy as np
import skops.io as sio
import torch
from datasets import load_dataset
from tqdm.auto import tqdm
from transformers import AutoModelForTokenClassification, HfArgumentParser

import adapters
import wtpsplit.models  # noqa: F401
from wtpsplit.evaluation import evaluate_mixture, get_labels, train_mixture
from wtpsplit.evaluation.intrinsic_baselines import split_language_data
from wtpsplit.extract import PyTorchWrapper, extract
from wtpsplit.models import SubwordXLMConfig, SubwordXLMForTokenClassification
from wtpsplit.utils import Constants, token_to_char_probs

logger = logging.getLogger()
logger.setLevel(logging.WARNING)


@dataclass
class Args:
    model_path: str
    adapter_path: str = None
    # eval data in the format:
    # {
    #    "<lang_code>": {
    #        "sentence": {
    #            "<dataset_name>": {
    #                 "meta": {
    #                     "train_data": ["train sentence 1", "train sentence 2"]
    #                 },
    #                 "data": ["test sentence 1", "test sentence 2"]
    #            }
    #        }
    #    }
    # }
    eval_data_path: str = "data/all_data.pth"
    valid_text_path: str = None  # "data/sentence/valid.parquet"
    device: str = "cpu"
    block_size: int = 512
    stride: int = 64
    batch_size: int = 32
    include_langs: List[str] = None
    custom_language_list: str = None
    threshold: float = 0.01
    max_n_train_sentences: int = 10000
    max_n_test_sentences: int = -1  # -1 is all
    keep_logits: bool = False
    skip_adaptation: bool = False
    skip_punct: bool = True
    skip_corrupted: bool = False
    clf_from_scratch: bool = False  # for FT + LoRA
    return_indices: bool = True
    exclude_every_k: int = 10
    save_suffix: str = ""
    num_hidden_layers: Union[int, None] = None  # for original XLM-R


def process_logits(text, model, lang_code, args):
    # Extract necessary data
    if isinstance(text, list):
        logits = []
        for short_seq in tqdm(text, desc="Listwise", disable=False):
            current_logits, current_offsets_mapping, tokenizer, _ = extract(
                [short_seq],
                model,
                lang_code=lang_code,
                stride=args.stride,
                max_block_size=args.block_size,
                batch_size=args.batch_size,
                pad_last_batch=True,
                verbose=False,
            )
            current_logits = current_logits[0]
            if current_offsets_mapping is not None:
                current_offsets_mapping = current_offsets_mapping[0]

            if "xlm" in model.config.model_type:
                tokens = tokenizer.tokenize(short_seq, verbose=False)

                char_probs = token_to_char_probs(short_seq, tokens, current_logits, tokenizer, current_offsets_mapping)

                current_logits = char_probs

            logits.append(current_logits)
    else:
        logits, offsets_mapping, tokenizer, _ = extract(
            [text],
            model,
            lang_code=lang_code,
            stride=args.stride,
            max_block_size=args.block_size,
            batch_size=args.batch_size,
            pad_last_batch=True,
            verbose=False,
        )
        logits = logits[0]
        if offsets_mapping is not None:
            offsets_mapping = offsets_mapping[0]

        if "xlm" in model.config.model_type:
            tokens = tokenizer.tokenize(text, verbose=False)

            # convert token probabilities to character probabilities for the entire array
            char_probs = token_to_char_probs(text, tokens, logits, tokenizer, offsets_mapping)

            logits = char_probs

    return logits


def load_or_compute_logits(args, model, eval_data, valid_data=None, save_str: str = None):
    logits_path = Constants.CACHE_DIR / "intrinsic" / f"{save_str}.h5"

    if not os.path.exists(Constants.CACHE_DIR / "intrinsic"):
        os.makedirs(Constants.CACHE_DIR / "intrinsic")

    if args.custom_language_list is not None:
        with open(args.custom_language_list, "r") as f:
            # file is a csv: l1,l2,...
            use_langs = f.read().strip().split(",")
    else:
        use_langs = eval_data.keys()

    total_test_time = 0

    with h5py.File(logits_path, "a") as f, torch.no_grad():
        for lang_code in tqdm(use_langs, desc="Languages"):
            if args.include_langs is not None and lang_code not in args.include_langs:
                continue

            if lang_code not in f:
                lang_group = f.create_group(lang_code)
            else:
                lang_group = f[lang_code]

            # valid data
            if valid_data is not None and "valid" not in lang_group:
                if args.adapter_path:
                    raise NotImplementedError("Adapters not supported for valid data")
                valid_sentences = [sample["text"].strip() for sample in valid_data if sample["lang"] == lang_code]
                assert len(valid_sentences) > 0

                separator = Constants.SEPARATORS.get(lang_code, " ")
                valid_text = separator.join(valid_sentences)

                valid_logits = process_logits(valid_text, model, lang_code, args)

                lang_group.create_dataset("valid", data=valid_logits)

            # eval data
            for dataset_name, dataset in tqdm(eval_data[lang_code]["sentence"].items(), desc=lang_code):
                if args.skip_corrupted and "corrupted" in dataset_name:
                    continue
                if "asr" in dataset_name and not any(
                    x in dataset_name for x in ["lyrics", "short", "code", "ted2020", "legal"]
                ):
                    logger.warning(f"SKIP: {lang_code} {dataset_name}")
                    continue
                if "legal" in dataset_name and not ("laws" in dataset_name or "judgements" in dataset_name):
                    logger.warning(f"SKIP: {lang_code} {dataset_name}")
                    continue
                if "social-media" in dataset_name:
                    logger.warning(f"SKIP: {lang_code} {dataset_name}")
                    continue
                if "nllb" in dataset_name:
                    continue
                if "-" in lang_code and "canine" in args.model_path and "no-adapters" not in args.model_path:
                    # code-switched data: eval 2x
                    lang_code = lang_code.split("_")[1].lower()
                try:
                    if args.adapter_path:
                        if args.clf_from_scratch:
                            model.model.classifier = torch.nn.Linear(model.model.classifier.in_features, 1)

                        dataset_load_name = dataset_name
                        model.model.load_adapter(
                            args.adapter_path + "/" + dataset_load_name + "/" + lang_code,
                            set_active=True,
                            with_head=True,
                            load_as="text",
                        )
                except Exception as e:
                    logger.error(f"Error loading adapter for {dataset_name} in {lang_code}: {e}")
                    continue
                if dataset_name not in lang_group:
                    dset_group = lang_group.create_group(dataset_name)
                else:
                    dset_group = lang_group[dataset_name]

                if "test_logits" not in dset_group:
                    test_sentences = dataset["data"]
                    if not test_sentences:
                        continue
                    max_n_sentences = args.max_n_test_sentences
                    test_sentences = test_sentences[:max_n_sentences]
                    if isinstance(test_sentences[0], list):
                        # short-seq eval: list of lists
                        test_text = [
                            Constants.SEPARATORS.get(lang_code, " ").join(sentence) for sentence in test_sentences
                        ]
                    else:
                        test_text = Constants.SEPARATORS.get(lang_code, " ").join(test_sentences)

                    start_time = time.time()
                    test_logits = process_logits(test_text, model, lang_code, args)
                    end_time = time.time()
                    total_test_time += end_time - start_time
                    if isinstance(test_sentences[0], list):
                        test_logit_lengths = []
                        # store start and end indices for each pair, used later to slice the logits
                        # (h5py does not like different length np arrays as list elements)
                        all_logit_lengths = np.append(0, np.cumsum([len(logits) for logits in test_logits]))
                        # append tuple of start and end indices for each pair
                        for i in range(len(test_logits)):
                            test_logit_lengths.append((all_logit_lengths[i], all_logit_lengths[i + 1] - 1))
                        test_logits = np.concatenate(test_logits)
                        # NOTE: handled differently than in intrinsic_pairwise.py
                        # here, we keep the label at the end
                        # in intrinsic_pairwise.py, we only consider non-ending labels.
                        test_labels = [
                            get_labels(lang_code, short_seq, after_space=False)[:-1] for short_seq in test_sentences
                        ]

                        # flatten; append 0 eos to account for later indexing/slicing
                        test_labels = np.append(np.concatenate(test_labels), 1)
                        assert len(test_labels) == len(test_logits) + 1
                        dset_group.create_dataset("test_logit_lengths", data=test_logit_lengths)
                    else:
                        test_labels = get_labels(lang_code, test_sentences, after_space=False)
                    if args.skip_punct:
                        # remove punct logits
                        test_logits = test_logits[:, 0]
                        # back to [N, 1]
                        test_logits = np.expand_dims(test_logits, axis=1)
                    dset_group.create_dataset("test_logits", data=test_logits)
                    dset_group.create_dataset("test_labels", data=test_labels)

                train_sentences = dataset["meta"].get("train_data")
                if train_sentences is not None and "train_logits" not in dset_group and not args.skip_adaptation:
                    train_sentences = train_sentences[: args.max_n_train_sentences]
                    if isinstance(train_sentences[0], list):
                        # short-seq eval: list of lists
                        train_text = [
                            Constants.SEPARATORS.get(lang_code, " ").join(sentence) for sentence in train_sentences
                        ]
                    else:
                        train_text = Constants.SEPARATORS.get(lang_code, " ").join(train_sentences)

                    train_logits = process_logits(train_text, model, lang_code, args)
                    if isinstance(train_sentences[0], list):
                        train_logits = np.concatenate(train_logits)
                        train_labels = [
                            get_labels(lang_code, short_seq, after_space=False)[:-1] for short_seq in train_sentences
                        ]

                        # flatten; append 0 eos to account for later indexing/slicing
                        train_labels = np.append(np.concatenate(train_labels), 1)
                        assert len(train_labels) == len(train_logits) + 1
                    else:
                        train_labels = get_labels(lang_code, train_sentences, after_space=False)

                    if args.skip_punct:
                        # remove punct logits
                        train_logits = train_logits[:, 0]
                        # back to [N, 1]
                        train_logits = np.expand_dims(train_logits, axis=1)
                    dset_group.create_dataset("train_logits", data=train_logits)
                    dset_group.create_dataset("train_labels", data=train_labels)

    end_time = time.time()
    return h5py.File(logits_path, "r"), total_test_time / 60  # to minutes


def compute_statistics(values):
    if not values:  # Check for empty values list
        return {"mean": None, "median": None, "std": None, "min": None, "min_lang": None, "max": None, "max_lang": None}

    scores, langs = zip(*values)
    min_index = np.argmin(scores)
    max_index = np.argmax(scores)
    return {
        "mean": np.mean(scores),
        "median": np.median(scores),
        "std": np.std(scores),
        "min": scores[min_index],
        "min_lang": langs[min_index],
        "max": scores[max_index],
        "max_lang": langs[max_index],
    }


def main(args):
    save_model_path = args.model_path
    if args.adapter_path:
        save_model_path = args.adapter_path
    save_str = f"{save_model_path.replace('/','_')}_b{args.block_size}_s{args.stride}"

    eval_data = torch.load(args.eval_data_path)
    if "canine" in args.model_path and "no-adapters" not in args.model_path:
        eval_data = split_language_data(eval_data)
    if args.valid_text_path is not None:
        valid_data = load_dataset("parquet", data_files=args.valid_text_path, split="train")
    else:
        valid_data = None

    logger.warning("Loading model...")
    model_path = args.model_path
    if args.model_path == "xlm-roberta-base" or args.model_path == "xlm-roberta-large":
        # init models here
        config = SubwordXLMConfig.from_pretrained(
            args.model_path,
            num_hidden_layers=args.num_hidden_layers,
            num_labels=1,
        )
        model = PyTorchWrapper(
            SubwordXLMForTokenClassification.from_pretrained(model_path, config=config).to(args.device)
        )
    else:
        model = PyTorchWrapper(AutoModelForTokenClassification.from_pretrained(model_path).to(args.device))
    if args.adapter_path:
        model_type = model.model.config.model_type
        # adapters need xlm-roberta as model type.
        model.model.config.model_type = "xlm-roberta"
        adapters.init(model.model)
        # reset model type (used later)
        model.model.config.model_type = model_type

    save_str += f"{args.save_suffix}"
    if args.max_n_test_sentences < sys.maxsize and args.max_n_test_sentences != -1:
        save_str += f"_n{args.max_n_test_sentences}"
    if args.max_n_test_sentences == -1:
        args.max_n_test_sentences = sys.maxsize

    # first, logits for everything.
    f, total_test_time = load_or_compute_logits(args, model, eval_data, valid_data, save_str)

    save_str += f"_u{args.threshold}"
    if args.exclude_every_k > 0 or "lyrics" in args.eval_data_path:
        save_str += f"_k{args.exclude_every_k}"

    # now, compute the intrinsic scores.
    results = {}
    clfs = {}
    if args.return_indices:
        indices = {}

    u_scores, t_scores, punct_scores = [], [], []

    for lang_code, dsets in tqdm(eval_data.items()):
        if args.include_langs is not None and lang_code not in args.include_langs:
            continue

        logger.warning(f"Predicting {lang_code}...")
        results[lang_code] = {}
        clfs[lang_code] = {}
        if args.return_indices:
            indices[lang_code] = {}

        for dataset_name, dataset in dsets["sentence"].items():
            sentences = dataset["data"]
            if not sentences:
                continue
            max_n_sentences = args.max_n_test_sentences
            sentences = sentences[:max_n_sentences]
            if len(sentences) == 0:
                continue
            if lang_code not in f or dataset_name not in f[lang_code]:
                continue

            # to be in line w/ LLM eval; for fair comparison
            if "lyrics" in dataset_name or "short" in dataset_name:
                exclude_every_k = 0
            else:
                exclude_every_k = args.exclude_every_k

            if "train_logits" in f[lang_code][dataset_name] and not args.skip_adaptation:
                feature_indices = None
                clf = train_mixture(
                    [lang_code],
                    f[lang_code][dataset_name]["train_logits"][:],
                    f[lang_code][dataset_name]["train_labels"][:],
                    features=feature_indices,
                    skip_punct=args.skip_punct,
                )
                if clf[0] is not None:
                    logger.warning(clf)

                if isinstance(sentences[0], list):
                    acc_t, acc_punct = [], []
                    score_t, score_punct = [], []
                    t_indices, punct_indices = [], []
                    for i, short_seq in enumerate(sentences):
                        start, end = f[lang_code][dataset_name]["test_logit_lengths"][i]
                        single_score_t, single_score_punct, info, cur_t_indices, cur_punct_indices = evaluate_mixture(
                            lang_code,
                            f[lang_code][dataset_name]["test_logits"][:][start:end],
                            list(short_seq),
                            args.return_indices,
                            exclude_every_k,
                            *clf,
                        )
                        score_t.append(single_score_t)
                        score_punct.append(single_score_punct)
                        acc_t.append(info["info_newline"]["correct_pairwise"] if info["info_newline"] else None)
                        acc_punct.append(
                            info["info_transformed"]["correct_pairwise"] if info["info_transformed"] else None
                        )
                        # indices: accumulate from start
                        t_indices.append(
                            cur_t_indices["pred_indices"] if cur_t_indices and cur_t_indices["pred_indices"] else []
                        )
                        punct_indices.append(
                            cur_punct_indices["pred_indices"]
                            if cur_punct_indices and cur_punct_indices["pred_indices"]
                            else []
                        )

                else:
                    score_t, score_punct, _, t_indices, punct_indices = evaluate_mixture(
                        lang_code,
                        f[lang_code][dataset_name]["test_logits"][:],
                        sentences,
                        args.return_indices,
                        exclude_every_k,
                        *clf,
                    )

                clfs[lang_code][dataset_name] = clf

                clf = list(copy.deepcopy(clf))
                clf[-1] = args.threshold
            else:
                score_t = score_punct = None
                clf = [None, None, None, args.threshold]
                t_indices, punct_indices = None, None

            if isinstance(sentences[0], list):
                acc_u = []
                score_u = []
                u_indices, true_indices = [], []
                length = []
                for i, short_seq in enumerate(sentences):
                    start, end = f[lang_code][dataset_name]["test_logit_lengths"][i]
                    single_score_u, _, info, cur_u_indices, _ = evaluate_mixture(
                        lang_code,
                        f[lang_code][dataset_name]["test_logits"][:][start:end],
                        list(short_seq),
                        args.return_indices,
                        exclude_every_k,
                        *clf,
                    )
                    score_u.append(single_score_u)
                    acc_u.append(info["info_newline"]["correct_pairwise"])
                    # indices: accumulate from start
                    u_indices.append(cur_u_indices["pred_indices"] if cur_u_indices["pred_indices"] else [])
                    true_indices.append(cur_u_indices["true_indices"] if cur_u_indices["true_indices"] else [])
                    length.append(cur_u_indices["length"])

            else:
                score_u, _, _, u_indices, _ = evaluate_mixture(
                    lang_code,
                    f[lang_code][dataset_name]["test_logits"][:],
                    sentences,
                    args.return_indices,
                    exclude_every_k,
                    *clf,
                )

            if isinstance(sentences[0], list):
                score_u = np.mean(score_u)
                score_t = np.mean(score_t) if score_t and not args.skip_adaptation else None
                score_punct = (
                    np.mean(score_punct) if score_punct and not (args.skip_punct or args.skip_adaptation) else None
                )
                acc_u = np.mean(acc_u)
                acc_t = np.mean(acc_t) if score_t else None
                acc_punct = np.mean(acc_punct) if score_punct else None

                results[lang_code][dataset_name] = {
                    "u": score_u,
                    "t": score_t,
                    "punct": score_punct,
                    "acc_u": acc_u,
                    "acc_t": acc_t,
                    "acc_punct": acc_punct,
                }
            else:
                results[lang_code][dataset_name] = {
                    "u": score_u,
                    "t": score_t,
                    "punct": score_punct,
                }

            if args.return_indices:
                if isinstance(sentences[0], list):
                    indices[lang_code][dataset_name] = {
                        "u": {"predicted_indices": u_indices, "true_indices": true_indices, "length": length},
                        "t": {"predicted_indices": t_indices, "true_indices": t_indices, "length": length}
                        if t_indices
                        else None,
                        "punct": {"predicted_indices": punct_indices, "true_indices": t_indices, "length": length}
                        if punct_indices
                        else None,
                    }
                else:
                    indices[lang_code][dataset_name] = {
                        "u": {
                            "predicted_indices": [u_indices["pred_indices"]],
                            "true_indices": [u_indices["true_indices"]],
                            "length": [u_indices["length"]],
                        },
                        "t": {
                            "predicted_indices": [t_indices["pred_indices"]],
                            "true_indices": [t_indices["true_indices"]],
                            "length": [t_indices["length"]],
                        }
                        if t_indices is not None
                        else None,
                        "punct": {
                            "predicted_indices": [punct_indices["pred_indices"]],
                            "true_indices": [punct_indices["true_indices"]],
                            "length": [punct_indices["length"]],
                        }
                        if punct_indices is not None
                        else None,
                    }

            if score_u is not None:
                u_scores.append((score_u, lang_code))
            if score_t is not None:
                t_scores.append((score_t, lang_code))
            if score_punct is not None:
                punct_scores.append((score_punct, lang_code))

            # just for logging
            score_t = score_t or 0.0
            score_punct = score_punct or 0.0
            logger.warning(f"{lang_code} {dataset_name} {score_u:.3f} {score_t:.3f} {score_punct:.3f}")

    # Compute statistics for each metric across all languages
    results_avg = {
        "u": compute_statistics(u_scores),
        "t": compute_statistics(t_scores),
        "punct": compute_statistics(punct_scores),
        "include_langs": args.include_langs,
    }

    if not args.skip_adaptation:
        sio.dump(
            clfs,
            open(
                Constants.CACHE_DIR / "intrinsic" / f"{save_str}.skops",
                "wb",
            ),
        )
    json.dump(
        results,
        open(
            Constants.CACHE_DIR / "intrinsic" / f"{save_str}.json",
            "w",
        ),
        indent=4,
    )
    print(Constants.CACHE_DIR / "intrinsic" / f"{save_str}.json")

    json.dump(
        results_avg,
        open(
            Constants.CACHE_DIR / "intrinsic" / f"{save_str}_AVG.json",
            "w",
        ),
        indent=4,
    )
    if args.return_indices:
        json.dump(
            indices,
            open(
                Constants.CACHE_DIR / "intrinsic" / f"{save_str}_IDX.json",
                "w",
            ),
            default=int,
            indent=4,
        )
        print(Constants.CACHE_DIR / "intrinsic" / f"{save_str}_IDX.json")
        print("Indices saved to file.")

    if not args.keep_logits:
        os.remove(f.filename)
    return results, results_avg, total_test_time


if __name__ == "__main__":
    (args,) = HfArgumentParser([Args]).parse_args_into_dataclasses()
    results, results_avg, total_test_time = main(args)
    print(total_test_time)



from dataclasses import dataclass

import pandas as pd
import regex as re
import sacrebleu
import skops.io as sio
from datasets import load_dataset
from tqdm.auto import tqdm
from transformers import AutoModelForTokenClassification, HfArgumentParser, pipeline

from wtpsplit.evaluation import ERSATZ_LANGUAGES, ersatz_sentencize, our_sentencize, preprocess_sentence
from wtpsplit.utils import Constants


@dataclass
class Args:
    model_path: str
    device: str = "cuda"
    clfs_path: str = Constants.CACHE_DIR / "main_3layers_h0_clfs.pkl"
    paragraph_length: int = 10


def mt_eval(
    sentence_model,
    lang_code,
    paragraph_length,
    clf,
    features,
    threshold_transformed,
    threshold_newline,
):
    dset_name = Constants.LANGINFO.loc[lang_code, "opus100"]
    dset_langs = dset_name.split("-")
    target_lang = dset_langs[0] if dset_langs[1] == lang_code else dset_langs[1]

    dset = load_dataset("opus100", dset_name)

    try:
        mt_pipeline = pipeline(
            "translation",
            model=f"Helsinki-NLP/opus-mt-{lang_code}-{target_lang}",
            device=0,
        )
    except OSError:
        return (None, None, None, None)

    all_sentences = [x[lang_code] for x in dset["test"]["translation"]]
    all_targets = [x[target_lang] for x in dset["test"]["translation"]]

    separator = Constants.SEPARATORS[lang_code]

    current_paragraph = []
    current_target = []

    all_preds_ground_truth = []
    all_preds_ersatz = []
    all_preds_ours = []
    all_preds_naive = []
    all_preds_no_sentencize = []

    all_target_texts = []

    bar = tqdm(total=len(all_sentences))

    while len(all_sentences) > 0:
        while len(current_paragraph) < paragraph_length and len(all_sentences) > 0:
            current_paragraph.append(preprocess_sentence(all_sentences.pop(0)))
            current_target.append(all_targets.pop(0))

            bar.update(1)

        paragraph_text = separator.join(current_paragraph)
        target_text = separator.join(current_target)

        preds_ground_truth = separator.join(
            [x["translation_text"] for x in mt_pipeline(current_paragraph, truncation=True)]
        )

        if lang_code in ERSATZ_LANGUAGES:
            preds_ersatz = separator.join(
                [
                    x["translation_text"]
                    for x in mt_pipeline(
                        [x.strip() for x in ersatz_sentencize(lang_code, paragraph_text)],
                        truncation=True,
                    )
                ]
            )
        else:
            preds_ersatz = None

        preds_ours = separator.join(
            [
                x["translation_text"]
                for x in mt_pipeline(
                    [
                        x.strip()
                        for x in our_sentencize(
                            sentence_model,
                            lang_code,
                            paragraph_text,
                            clf,
                            features,
                            threshold_transformed,
                            threshold_newline,
                        )
                    ],
                    truncation=True,
                )
            ]
        )
        preds_no_sentencize = mt_pipeline(paragraph_text, truncation=True)[0]["translation_text"]
        if Constants.LANGINFO.loc[lang_code, "no_whitespace"]:
            naive_sentences = [
                paragraph_text[
                    int(i * len(paragraph_text) / len(current_paragraph)) : int(
                        (i + 1) * len(paragraph_text) / len(current_paragraph)
                    )
                ]
                for i in range(len(current_paragraph))
            ]
        else:
            words = [x for x in re.split(r"(\S+\s+)", paragraph_text) if len(x) > 0]
            naive_sentences = [
                "".join(
                    words[
                        int(i * len(words) / len(current_paragraph)) : int(
                            (i + 1) * len(words) / len(current_paragraph)
                        )
                    ]
                )
                for i in range(len(current_paragraph))
            ]

        assert "".join(naive_sentences) == paragraph_text
        preds_naive = separator.join(
            [x["translation_text"] for x in mt_pipeline([x.strip() for x in naive_sentences], truncation=True)]
        )

        all_preds_naive.append(preds_naive)
        all_preds_ground_truth.append(preds_ground_truth)
        if preds_ersatz is not None:
            all_preds_ersatz.append(preds_ersatz)
        all_preds_ours.append(preds_ours)
        all_preds_no_sentencize.append(preds_no_sentencize)

        all_target_texts.append(target_text)

        current_paragraph = []
        current_target = []

    bleu = sacrebleu.BLEU()

    return (
        bleu.corpus_score(all_preds_no_sentencize, [all_target_texts]).score,
        bleu.corpus_score(all_preds_ground_truth, [all_target_texts]).score,
        bleu.corpus_score(all_preds_naive, [all_target_texts]).score,
        bleu.corpus_score(all_preds_ersatz, [all_target_texts]).score if len(all_preds_ersatz) > 0 else None,
        bleu.corpus_score(all_preds_ours, [all_target_texts]).score,
    )


if __name__ == "__main__":
    (args,) = HfArgumentParser([Args]).parse_args_into_dataclasses()

    model = AutoModelForTokenClassification.from_pretrained(args.model_path).to(args.device)

    clfs = sio.load(open(args.clfs_path, "rb"), ["numpy.float64", "numpy.float32"])

    langs = [
        "ar",
        "cs",
        "de",
        "en",
        "es",
        "fi",
        "hi",
        "ja",
        "ka",
        "lv",
        "pl",
        "th",
        "xh",
        "zh",
    ]

    results = []

    for lang_code in tqdm(langs):
        results.append(
            (
                lang_code,
                *mt_eval(
                    model,
                    lang_code,
                    args.paragraph_length,
                    *clfs[lang_code]["opus100"],
                ),
            ),
        )
        print(results[-1])

    pd.DataFrame(results, columns=["lang", "no_sent", "ground_truth", "naive", "ersatz", "ours"]).to_csv(
        "results.csv", index=False
    )



import logging
import math
import os
import random
import shutil
import sys

# import time
from collections import Counter, defaultdict
from dataclasses import dataclass
from functools import partial
from glob import glob
from typing import List, Optional

import datasets
import numpy as np
import torch
import transformers
from datasets import load_dataset

# from datasets.download import DownloadConfig
from tokenizers import AddedToken
from torchinfo import summary
from tqdm.auto import tqdm
from transformers import AutoTokenizer, HfArgumentParser, TrainingArguments, set_seed
from transformers.trainer import is_torch_tpu_available

import wandb
from wtpsplit.models import (
    BertCharConfig,
    BertCharForTokenClassification,
    LACanineConfig,
    LACanineForTokenClassification,
    SubwordXLMConfig,
    SubwordXLMForTokenClassification,
)
from wtpsplit.train.evaluate import evaluate_sentence
from wtpsplit.train.trainer import Trainer
from wtpsplit.train.utils import Model

# from wtpsplit.train.utils import cleanup_cache_files
from wtpsplit.utils import Constants, LabelArgs, corrupt_training, get_label_dict, get_subword_label_dict

logger = logging.getLogger(__name__)


def setup_logging(training_args: transformers.TrainingArguments) -> None:
    logging.basicConfig(
        format="%(asctime)s - %(levelname)s - %(name)s - %(message)s",
        datefmt="%m/%d/%Y %H:%M:%S",
        handlers=[logging.StreamHandler(sys.stdout)],
    )
    log_level = training_args.get_process_log_level()
    logger.setLevel(log_level)
    datasets.utils.logging.set_verbosity_warning()
    transformers.utils.logging.set_verbosity_warning()
    transformers.utils.logging.enable_default_handler()
    transformers.utils.logging.enable_explicit_format()
    # Log on each process the small summary:
    logger.warning(
        (
            f"Process rank: {training_args.local_rank}, device: {training_args.device}, n_gpu: {training_args.n_gpu}"
            + f"distributed training: {training_args.local_rank != -1}, 16-bits training: {training_args.fp16}"
        )
    )


@dataclass
class Args:
    model_name_or_path: str
    shuffle: bool = False
    use_logits: bool = False
    is_decoder: bool = False
    use_bert: bool = False
    train_text_path: str = "data/train.parquet"
    valid_text_path: str = "data/valid.parquet"
    include_languages: List[str] = None
    eval_data_path: str = "data/all_data.pth"
    num_hidden_layers: int = 3
    preprocessing_num_workers: int = 6
    block_size: int = 512
    overflow_size: int = 16
    eval_stride: int = 256
    loss_margin: float = 0.5
    from_scratch: bool = False
    pack_samples: bool = False
    one_sample_per_line: bool = False
    use_loss_weights: bool = False
    do_sentence_training: bool = True
    do_auxiliary_training: bool = True
    aux_training_weight: float = 1.0
    ignore_non_hyphen: bool = False
    non_punctuation_sample_ratio: float = None
    text_column: str = "text"
    threshold: float = 0.01  # just for eval
    # WtP-related args
    ngram_order: int = 1
    language_adapter: str = "on"
    adapter_warmup_steps: int = 0
    adapter_lr_multiplier: float = 1.0
    # SaT-related args
    use_subwords: bool = False  # uses XLM-R
    lookahead: int = None
    lookahead_split_layers: Optional[int] = None
    sample_non_whitespace: int = 1


def collate_fn(batch, args, label_args, label_dict, tokenizer, add_lang_ids: bool = False):
    all_input_ids = []
    all_labels = []
    all_language_ids = []

    all_attention_masks = []
    all_position_ids = []
    all_label_weights = []

    for sample in batch:
        # subword-level
        if args.use_subwords:
            # already tokenized!
            input_ids = sample["input_ids"]
        # char-level
        else:
            input_ids = [ord(c) for c in sample["input_ids"]]
        lang = sample["lang"]

        while len(input_ids) < args.block_size + args.overflow_size:
            if tokenizer:
                input_ids.append(tokenizer.pad_token_id)
            else:
                input_ids.append(0)

        block_ids = [0] * len(input_ids)

        input_ids, _, labels = corrupt_training(
            input_ids,
            block_ids,
            lang,
            label_args,
            label_dict=label_dict,
            pack_samples=args.pack_samples,
            # min_length=args.block_size,
            tokenizer=tokenizer if args.use_subwords else None,
        )

        actual_block_size = args.block_size - 2 if args.use_subwords else args.block_size

        if len(input_ids) > args.block_size:
            start = np.random.randint(0, len(input_ids) - actual_block_size)
            input_ids = input_ids[start : start + actual_block_size]
            labels = labels[start : start + actual_block_size]
        elif len(input_ids) < actual_block_size:
            padding = actual_block_size - len(input_ids)
            input_ids += [tokenizer.pad_token_id] * padding if tokenizer else [0] * padding
            labels += [0] * padding

        if tokenizer:
            input_ids = [tokenizer.cls_token_id] + input_ids[:actual_block_size] + [tokenizer.sep_token_id]
            # labels for CLS and SEP tokens are 0 (none)
            labels = [0] + labels[:actual_block_size] + [0]
        else:
            input_ids = input_ids[:actual_block_size]
            labels = labels[:actual_block_size]

        input_ids = torch.tensor(input_ids, dtype=torch.long)
        labels = torch.tensor(labels, dtype=torch.long)
        position_ids = torch.arange(len(input_ids), dtype=torch.long)
        label_weights = torch.ones(args.block_size, dtype=torch.float32)
        if tokenizer:
            attention_mask = (input_ids != tokenizer.pad_token_id).to(torch.float32)
        else:
            attention_mask = (input_ids != 0).to(torch.float32)

        all_input_ids.append(input_ids)
        all_label_weights.append(label_weights)
        all_labels.append(labels)
        all_language_ids.append(Constants.LANG_CODE_TO_INDEX[lang] if add_lang_ids else 0)

        all_attention_masks.append(attention_mask)
        all_position_ids.append(position_ids)

    out = {
        "input_ids": torch.stack(all_input_ids, 0),
        "attention_mask": torch.stack(all_attention_masks, 0),
        "position_ids": torch.stack(all_position_ids, 0),
        "language_ids": torch.tensor(all_language_ids, dtype=torch.long),
        "label_weights": torch.stack(all_label_weights, 0),
        "labels": torch.stack(all_labels, 0),
    }

    return out


def main():
    parser = HfArgumentParser([Args, TrainingArguments, LabelArgs])

    if sys.argv[1].endswith(".json"):
        (args, training_args, label_args) = parser.parse_json_file(sys.argv[1])
        wandb_name = os.path.splitext(os.path.basename(sys.argv[1]))[0]
    else:
        (args, training_args, label_args) = parser.parse_args_into_dataclasses()
        wandb_name = None

    if is_torch_tpu_available():
        import torch_xla.core.xla_model as xm

        world_size = xm.xrt_world_size()
        if world_size == 4:
            # ensure same batch size on TPUv3 and TPUv4 using same config.json
            training_args.per_device_train_batch_size *= 2
    elif torch.cuda.is_available():
        world_size = torch.cuda.device_count()
    else:
        world_size = 1

    logger.warning(f"Per device train batch size: {training_args.per_device_train_batch_size}")
    logger.warning(
        f"Total train batch size: {training_args.per_device_train_batch_size * training_args.gradient_accumulation_steps * world_size}"
    )

    setup_logging(training_args)
    set_seed(training_args.seed)
    training_args.hub_strategy = "end"
    training_args.save_total_limit = 1

    num_labels = Constants.AUX_OFFSET + ((1 + len(Constants.PUNCTUATION_CHARS)) if args.do_auxiliary_training else 0)
    if args.use_subwords:
        # SaT models
        if args.from_scratch:
            config = SubwordXLMConfig(
                args.model_name_or_path,
                num_hidden_layers=args.num_hidden_layers,
                num_labels=num_labels,
                lookahead=args.lookahead,
                lookahead_split_layers=args.lookahead_split_layers,
            )
            backbone = SubwordXLMForTokenClassification(config)

        else:
            config = SubwordXLMConfig.from_pretrained(
                args.model_name_or_path,
                num_hidden_layers=args.num_hidden_layers,
                num_labels=num_labels,
                lookahead=args.lookahead,
                lookahead_split_layers=args.lookahead_split_layers,
            )
            backbone = SubwordXLMForTokenClassification.from_pretrained(
                args.model_name_or_path,
                config=config,
            )

        backbone.config.base_model = args.model_name_or_path
        tokenizer = AutoTokenizer.from_pretrained(args.model_name_or_path)
        # needed since we create labels in collate_fn based on tokens
        tokenizer.add_special_tokens({"additional_special_tokens": [AddedToken("\n")]})
        custom_token_id = tokenizer.convert_tokens_to_ids("\n")
        # used later to filter out special tokens
        special_tokens_ids = set(tokenizer.all_special_ids)
        special_tokens_ids.discard(custom_token_id)
        if args.lookahead:
            # we split lookahead evenly into N layers
            assert args.lookahead % args.num_hidden_layers == 0

    else:
        # WtP models (char-based)
        tokenizer = None
        config = LACanineConfig.from_pretrained(
            args.model_name_or_path,
            raw_lookahead=args.lookahead,
            num_hidden_layers=args.num_hidden_layers,
            num_labels=num_labels,
            n_languages=len(Constants.LANG_CODE_TO_INDEX),
            ngram_order=args.ngram_order,
            language_adapter=args.language_adapter,
            # upsampling kernel size > 1 is problematic for packing
            # using ks=1 doesn't allow reusing the pretrained weights
            # but if we warm it up alongside the adapters
            # there is almost no difference.
            upsampling_kernel_size=1,
        )
        if args.use_bert:
            config = BertCharConfig.from_pretrained(
                args.model_name_or_path,
                num_labels=num_labels,
            )
            backbone = BertCharForTokenClassification(config)
        elif args.from_scratch:
            backbone = LACanineForTokenClassification(config)
        else:
            backbone = LACanineForTokenClassification.from_pretrained(
                args.model_name_or_path, ignore_mismatched_sizes=True, config=config
            )

    model = Model(
        backbone,
        loss_margin=args.loss_margin,
        use_loss_weights=args.use_loss_weights,
        do_sentence_training=args.do_sentence_training,
        do_auxiliary_training=args.do_auxiliary_training,
        aux_training_weight=args.aux_training_weight,
    )

    if training_args.local_rank == 0:
        logger.warning(summary(model, depth=4))

    def prepare_dataset(
        num_workers=1,
        include_languages=None,
        shuffle=False,
        split="train",
    ):
        with training_args.main_process_first():
            # this can be used if space issues arise
            # dlconf = DownloadConfig(cache_dir="/home/Markus/.cache/huggingface/datasets")
            # dataset = load_dataset("markus583/mC4-TEST", split=split, download_config=dlconf)
            dataset = load_dataset("markus583/mC4-TEST", split=split)
        logger.warning(f"Loaded {split} dataset.")
        # optional: delete downloaded dataset, it is stored in cache_dir now (but we delete it later)
        # ~40GB on disk
        # os.system("rm -rf /home/Markus/.cache/huggingface/datasets")

        if include_languages is not None:
            include_languages = set(include_languages)

            dataset = dataset.filter(
                lambda example: example["lang"] in include_languages,
                num_proc=args.preprocessing_num_workers,
            )
            logger.warning(f"Filtered to {len(dataset)} examples.")

        if shuffle:
            dataset = dataset.shuffle(seed=42)
            logger.warning("Shuffled dataset.")

        # not used for sentence segmentation, ignore.
        if args.ignore_non_hyphen:
            with training_args.main_process_first():
                dataset = dataset.filter(
                    lambda sample: any(c in sample[args.text_column] for c in label_args.hyphen_chars),
                    num_proc=args.preprocessing_num_workers,
                )
                logger.info(f"Filtered to {len(dataset)} examples.")

        # "punctuation-specific sampling" in the WtP paper
        if args.non_punctuation_sample_ratio is not None:
            languages_without_punctuation = {
                lang_code
                for lang_code in Constants.LANGINFO.index
                if Constants.LANGINFO.loc[lang_code, "no_punctuation"]
            }

            def drop_some_non_punctuation_samples(examples):
                include_indices = set(
                    np.where([lang_code not in languages_without_punctuation for lang_code in examples["lang"]])[0]
                )
                punctuation_indices = {
                    i for i in np.where(examples["ends_with_punctuation"])[0] if i in include_indices
                }

                target_n_non_punct = int(
                    (len(punctuation_indices) * args.non_punctuation_sample_ratio)
                    / (1 - args.non_punctuation_sample_ratio)
                )
                n_drop = (len(include_indices) - len(punctuation_indices)) - target_n_non_punct

                out = [True for _ in range(len(examples["ends_with_punctuation"]))]

                if n_drop <= 0:
                    return out
                drop_indices = np.random.choice(
                    list(include_indices - punctuation_indices),
                    n_drop,
                    replace=False,
                )

                for i in drop_indices:
                    out[i] = False

                return out

            with training_args.main_process_first():
                dataset = dataset.filter(
                    drop_some_non_punctuation_samples,
                    batched=True,
                    batch_size=1_000_000,
                    num_proc=num_workers,
                )

        def tokenize_texts(examples):
            # do not return CLS and SEP token here
            # there should only be 1 of these per block later, not multiple
            # we still can't use return_special_tokens=False since we need the \n token later for the labels
            tokenized = tokenizer(examples[args.text_column], verbose=False)
            return {"input_ids": [example[1:-1] for example in tokenized["input_ids"]]}

        # similar to group_texts in huggingface's run_clm.py / run_mlm.py: https://github.com/huggingface/transformers/blob/main/examples/pytorch/language-modeling/run_mlm.py
        def group_texts(examples):
            all_input_blocks = []
            all_input_block_lengths = []
            all_langs = []

            def maybe_pad(text):
                if args.pack_samples:
                    padding = config.downsampling_rate - (len(text) % config.downsampling_rate)
                    if padding == config.downsampling_rate:
                        padding = 0

                    text += chr(0) * padding

                return text

            for current_lang in set(examples["lang"]):
                if not args.use_subwords:
                    lang_texts = [
                        maybe_pad(text)
                        for text, lang in zip(examples["input_ids"], examples["lang"])
                        if lang == current_lang
                    ]
                else:
                    # only retain current_lang examples (all columns)
                    lang_subwords = [
                        subwords
                        for subwords, lang in zip(examples["input_ids"], examples["lang"])
                        if lang == current_lang
                    ]
                    # filter out some special tokens
                    # from html tags, mostly in Latin, Thai & Korean
                    lang_subwords = [
                        [subword for subword in subwords if subword not in special_tokens_ids]
                        for subwords in lang_subwords
                    ]
                # pack_samples used for the compound part, so irrelevant
                if args.pack_samples:
                    if args.use_subwords:
                        raise NotImplementedError
                    blocks = []
                    block_ids = []

                    current_block = ["", []]

                    for i, text in enumerate(lang_texts):
                        if len(text) > args.block_size:
                            continue

                        current_block[0] += text
                        current_block[1] += [i] * len(text)

                        if i + 1 < len(lang_texts) and len(current_block[0]) + len(lang_texts[i + 1]) > args.block_size:
                            padding = args.block_size - len(current_block[0])

                            current_block[0] += chr(0) * padding
                            current_block[1] += [i] * padding
                            blocks.append(current_block[0])
                            block_ids.append(current_block[1])

                            current_block = ["", []]

                    if len(current_block[0]) > 0:
                        padding = args.block_size - len(current_block[0])

                        current_block[0] += chr(0) * padding
                        current_block[1] += [i] * padding
                        blocks.append(current_block[0])
                        block_ids.append(current_block[1])
                else:
                    if not args.use_subwords:
                        concatenated_texts = "".join(lang_texts)
                        concatenated_ids = [i for i, text in enumerate(lang_texts) for _ in text]
                    else:
                        # concatenate token lists
                        concatenated_texts = [item for sublist in lang_subwords for item in sublist]
                        concatenated_ids = [i for i, subwords in enumerate(lang_subwords) for _ in subwords]

                    total_length = len(concatenated_texts)

                    best_length = math.ceil(total_length / args.block_size) * args.block_size + args.overflow_size
                    while best_length > total_length:
                        best_length -= args.block_size

                    if best_length < 0:
                        continue

                    concatenated_texts = concatenated_texts[:best_length]
                    concatenated_ids = concatenated_ids[:best_length]

                    blocks = [
                        concatenated_texts[i : i + args.block_size + args.overflow_size]
                        for i in range(0, best_length - args.block_size, args.block_size)
                    ]
                    block_ids = [
                        concatenated_ids[i : i + args.block_size + args.overflow_size]
                        for i in range(0, best_length - args.block_size, args.block_size)
                    ]

                block_langs = [current_lang] * len(blocks)

                all_input_blocks.extend(blocks)
                all_input_block_lengths.extend([list(Counter(ids).values()) for ids in block_ids])
                all_langs.extend(block_langs)

                if args.sample_non_whitespace > 1:
                    separator = Constants.SEPARATORS.get(current_lang, " ")
                    if separator == "":
                        for i in range(args.sample_non_whitespace - 1):
                            all_input_blocks.extend(blocks)
                            all_input_block_lengths.extend([list(Counter(ids).values()) for ids in block_ids])
                            all_langs.extend(block_langs)

            return {
                "input_ids": all_input_blocks,
                "block_lengths": all_input_block_lengths,
                "lang": all_langs,
            }

        if args.do_auxiliary_training:
            assert label_args.use_auxiliary

        if args.pack_samples:
            assert not args.one_sample_per_line

        if args.use_subwords:
            with training_args.main_process_first():
                dataset = dataset.map(
                    tokenize_texts,
                    batched=True,
                    num_proc=num_workers,
                    remove_columns=[args.text_column],
                )
        else:
            # this column is no longer used and would cause an error otherwise
            with training_args.main_process_first():
                dataset = dataset.rename_column(args.text_column, "input_ids")
        logger.warning(f"Tokenized {split} dataset.")

        # uncomment if space issues arise (e.g., on TPU VMs):
        # if split == "train" and args.use_subwords:
        #     with training_args.main_process_first():
        #         for root, dirs, files in os.walk(os.environ.get("HF_DATASETS_CACHE")):
        #             for file in files:
        #                 if file.startswith("m_c4-test-train"):
        #                     logger.warning(f"Removing {os.path.join(root, file)}")
        #                     os.remove(os.path.join(root, file))

        if not args.one_sample_per_line:
            with training_args.main_process_first():
                dataset = dataset.map(
                    group_texts,
                    batched=True,
                    num_proc=num_workers,
                    # a bit hacky but oh well, only drop if sentence
                    remove_columns=["ends_with_punctuation"] if args.text_column == "text" else [],
                )
        logger.warning(f"Grouped {split} dataset.")

        return dataset

    valid_dataset = prepare_dataset(
        num_workers=args.preprocessing_num_workers,
        include_languages=args.include_languages,
        shuffle=False,
        split="valid",
    )
    logger.warning(f"Valid dataset has {len(valid_dataset)} examples.")

    train_dataset = prepare_dataset(
        num_workers=args.preprocessing_num_workers,
        include_languages=args.include_languages,
        shuffle=args.shuffle,
        split="train",
    )
    logger.warning(f"Train dataset has {len(train_dataset)} examples.")

    # print some samples from the dataset
    count = 0
    while count < 5:
        index = random.choice(range(len(train_dataset)))
        sample = train_dataset[index]

        logger.warning(f"Sample {index} of the training set: {sample}.")
        if tokenizer:
            logger.warning(tokenizer.decode(sample["input_ids"]))
        count += 1

    eval_data = torch.load(
        args.eval_data_path,
    )

    def compute_metrics(trainer):
        metrics = {}
        avg_metrics = defaultdict(lambda: [])

        model = trainer._wrap_model(trainer.model, training=False)

        for lang_code, lang_data in tqdm(eval_data.items(), desc="Evaluate!"):
            if args.include_languages is not None and lang_code not in args.include_languages:
                continue

            if trainer.args.process_index == 0 and args.do_sentence_training:
                for dataset_name, dataset in lang_data["sentence"].items():
                    if not dataset["data"][0]:
                        continue

                    if isinstance(dataset["data"][0], list):
                        # too slow here
                        continue
                    score, info = evaluate_sentence(
                        lang_code,
                        dataset["data"],
                        model,
                        stride=args.eval_stride,
                        block_size=args.block_size,
                        batch_size=training_args.per_device_eval_batch_size,
                        threshold=args.threshold,
                    )
                    metrics[f"{lang_code}_{dataset_name}_pr_auc"] = score
                    metrics[f"{lang_code}_{dataset_name}_f1"] = info["f1"]
                    metrics[f"{lang_code}_{dataset_name}_f1_best"] = info["f1_best"]
                    metrics[f"{lang_code}_{dataset_name}_threshold_best"] = info["threshold_best"]
                    avg_metrics[f"average_{dataset_name}_pr_auc"].append(score)
                    avg_metrics[f"average_{dataset_name}_f1"].append(info["f1"])
                    avg_metrics[f"average_{dataset_name}_f1_best"].append(info["f1_best"])
                    avg_metrics[f"average_{dataset_name}_threshold_best"].append(info["threshold_best"])

                    if lang_code in ["zh", "ja", "my", "km"]:
                        avg_metrics[f"average_nonwhitespace_{dataset_name}_pr_auc"].append(score)
                        avg_metrics[f"average_nonwhitespace_{dataset_name}_f1"].append(info["f1"])
                        avg_metrics[f"average_nonwhitespace_{dataset_name}_f1_best"].append(info["f1_best"])
                        avg_metrics[f"average_nonwhitespace_{dataset_name}_threshold_best"].append(
                            info["threshold_best"]
                        )
                    else:
                        avg_metrics[f"average_whitespace_{dataset_name}_pr_auc"].append(score)
                        avg_metrics[f"average_whitespace_{dataset_name}_f1"].append(info["f1"])
                        avg_metrics[f"average_whitespace_{dataset_name}_f1_best"].append(info["f1_best"])
                        avg_metrics[f"average_whitespace_{dataset_name}_threshold_best"].append(info["threshold_best"])

        for name, values in avg_metrics.items():
            if len(values) > 1:
                metrics[name] = np.mean(values)

        return metrics

    if "wandb" in training_args.report_to and training_args.process_index == 0:
        wandb.init(name=wandb_name, project="sentence")
        wandb.config.update(args)
        wandb.config.update(training_args)
        wandb.config.update(label_args)

        model.config.wandb_run_id = wandb.run.id

        for file in glob(os.path.join(os.path.dirname(__file__), "*.py")):
            wandb.save(os.path.abspath(file), policy="now")
            # also 1 dir above
            wandb.save(os.path.abspath(os.path.join(os.path.dirname(__file__), "..", file)), policy="now")

    label_dict = get_subword_label_dict(label_args, tokenizer) if args.use_subwords else get_label_dict(label_args)
    logger.info(f"Label dict has {len(label_dict)} entries.")

    # needed in the trainer
    training_args.adapter_warmup_steps = args.adapter_warmup_steps
    training_args.adapter_lr_multiplier = args.adapter_lr_multiplier

    # again: uncomment this if space issues arise.
    # give .map in multiprocessing enough of time to finish, to be safe
    # time.sleep(10)
    # if training_args.local_rank == 0:
    #     # since both share the *same* cache_dir, we cannot simply call dataset.cleanup_cache_files()
    #     # because that would remove the cache files of the other dataset!
    #     cleanup_cache_files([train_dataset, valid_dataset])
    #     logger.warning("Cleaned up cache files.")
    # time.sleep(10)

    trainer = Trainer(
        model,
        training_args,
        train_dataset=train_dataset,
        eval_dataset=valid_dataset,
        compute_metrics=compute_metrics,
        data_collator=partial(
            collate_fn,
            args=args,
            label_args=label_args,
            label_dict=label_dict,
            tokenizer=tokenizer if args.use_subwords else None,
            add_lang_ids=not args.use_subwords,
        ),
    )

    trainer.train(resume_from_checkpoint=training_args.resume_from_checkpoint)
    trainer.save_model()
    trainer.save_state()

    # remove old checkpoints to save space
    checkpoint_pattern = os.path.join(training_args.output_dir, "checkpoint-*")

    for checkpoint_dir in glob(checkpoint_pattern):
        if os.path.isdir(checkpoint_dir):
            shutil.rmtree(checkpoint_dir)


def _mp_fn(index):
    # For xla_spawn (TPUs)
    main()


if __name__ == "__main__":
    main()



import logging
import sys

import numpy as np
import pysbd
import sklearn.metrics

from wtpsplit.evaluation.intrinsic_pairwise import generate_k_mers, process_logits_k_mers
from wtpsplit.extract import PyTorchWrapper, extract
from wtpsplit.utils import Constants, sigmoid, corrupt, token_to_char_probs

logger = logging.getLogger(__name__)


def compute_iou(a, b):
    return len(set(a) & set(b)) / len(set(a) | set(b))


def compute_f1(pred, true):
    pred = set(pred)
    true = set(true)

    tp = len(true & pred)
    fp = len(pred - true)
    fn = len(true - pred)

    precision = tp / (tp + fp)
    recall = tp / (tp + fn)

    f1 = 2 * (precision * recall) / (precision + recall)

    return (
        f1,
        precision,
    )


def get_metrics(labels, preds, threshold: float = 0.01):
    # Compute precision-recall curve and AUC
    precision, recall, thresholds = sklearn.metrics.precision_recall_curve(labels, sigmoid(preds))
    pr_auc = sklearn.metrics.auc(recall, precision)

    # Compute F1 scores for all thresholds
    f1_scores = 2 * (precision * recall) / (precision + recall + 1e-10)

    # Find best F1 score and its corresponding threshold
    best_f1_index = np.argmax(f1_scores[:-1])  # Exclude last value because it corresponds to recall of 0.
    best_f1 = f1_scores[best_f1_index]
    best_threshold = thresholds[best_f1_index]

    # Compute F1 score for a specific threshold (e.g., 0.01 after applying sigmoid)
    f1_at_specific_threshold = sklearn.metrics.f1_score(labels, sigmoid(preds) > threshold)

    metrics = {"pr_auc": pr_auc}
    info = {
        "probs": preds,
        "labels": labels,
        "recalls": recall,
        "precisions": precision,
        "f1_scores": f1_scores,
        "f1_best": best_f1,
        "threshold_best": sigmoid(best_threshold).item(),
        "f1": f1_at_specific_threshold,
        "thresholds": thresholds,
    }

    return metrics, info


def evaluate_sentence(
    lang_code,
    sentences,
    model,
    stride,
    block_size,
    batch_size,
    use_pysbd=False,
    positive_index=None,
    do_lowercase=False,
    do_remove_punct=False,
    threshold: float = 0.01,
):
    if positive_index is None:
        positive_index = Constants.NEWLINE_INDEX

    # preprocessing, many OPUS100 (and some UD) sentences start with "- "
    sentences = [sentence.lstrip("-").strip() for sentence in sentences]

    separator = Constants.SEPARATORS[lang_code]
    sentences = [corrupt(sentence, do_lowercase, do_remove_punct) for sentence in sentences]
    text = separator.join(sentences)

    logits, offsets_mapping, tokenizer, _ = extract(
        [text],
        PyTorchWrapper(model.backbone),
        lang_code=lang_code,
        stride=stride,
        max_block_size=block_size,
        batch_size=batch_size,
    )
    logits = logits[0]
    if offsets_mapping is not None:
        offsets_mapping = offsets_mapping[0]

    true_end_indices = np.cumsum(np.array([len(s) for s in sentences])) + np.arange(len(sentences)) * len(separator)
    newline_labels = np.zeros(len(text))
    newline_labels[true_end_indices - 1] = 1

    if "xlm" in model.config.model_type:
        tokens = tokenizer.tokenize(text, verbose=False)
        char_probs = token_to_char_probs(text, tokens, logits, tokenizer, offsets_mapping)
    else:
        char_probs = logits
    newline_probs = char_probs[:, positive_index]
    metrics, info = get_metrics(newline_labels, newline_probs, threshold=threshold)

    info["newline_labels"] = newline_labels

    if use_pysbd:
        segmenter = pysbd.Segmenter(language=lang_code, clean=False, char_span=True)
        predicted_end_indices_pysbd = np.array([x.start + len(x.sent.rstrip()) for x in segmenter.segment(text)])
        newline_probs_pysbd = np.zeros(len(text))
        newline_probs_pysbd[predicted_end_indices_pysbd - 1] = 1.0

        info["newline_probs_pysbd"] = newline_probs_pysbd

    return metrics["pr_auc"], info


def evaluate_sentence_pairwise(
    lang_code,
    sentences,
    model,
    stride,
    block_size,
    batch_size,
    pair_sample_pct: float = 0.1,
    max_pairs: int = sys.maxsize,
    use_pysbd=False,
    positive_index=None,
    do_lowercase=False,
    do_remove_punct=False,
    threshold: float = 0.01,
):
    if positive_index is None:
        positive_index = Constants.NEWLINE_INDEX

    # Preprocess sentences
    sentences = [sentence.lstrip("-").strip() for sentence in sentences]

    separator = Constants.SEPARATORS[lang_code]
    metrics_list = []
    accuracy_list = []

    # get pairs of sentences (non-overlapping)
    sampled_pairs = generate_k_mers(
        sentences=sentences,
        k=2,
        sample_pct=pair_sample_pct,
        max_n_samples=max_pairs,
        min_k_mer_length=0,
    )

    # get logits for each pair
    logits, n_tokens_list = process_logits_k_mers(
        pairs=sampled_pairs,
        model=PyTorchWrapper(model.backbone),
        lang_code=lang_code,
        block_size=block_size,
        batch_size=batch_size,
        verbose=False,
    )

    # simulate performance for WtP-U
    for i, (sentence1, sentence2) in enumerate(sampled_pairs):
        newline_probs = logits[i][:, positive_index]

        pair_text = sentence1 + separator + sentence2

        # Calculate newline labels and probabilities
        true_end_indices = np.cumsum(np.array([len(s) for s in [sentence1, sentence2]])) + np.arange(2) * len(separator)
        newline_labels = np.zeros(len(pair_text))
        newline_labels[true_end_indices - 1] = 1

        # Get metrics for the pair
        pair_metrics, _ = get_metrics(newline_labels, newline_probs, threshold=threshold)
        metrics_list.append(pair_metrics["pr_auc"])
        predicted_labels = newline_probs > np.log(threshold / (1 - threshold))  # inverse sigmoid
        # for accuracy, check if the single label in between is correctly predicted (ignore the one at the end)
        if sum(predicted_labels[:-1]) > 0:
            correct = (np.where(newline_labels[:-1])[0] == np.where(predicted_labels[:-1])[0]).all()
            accuracy_list.append(correct)
        else:
            accuracy_list.append(False)

    # Compute and return the average metric
    average_metric = np.mean(metrics_list)
    avg_accuracy = np.mean(accuracy_list)
    return average_metric, avg_accuracy


def evaluate_sentence_kmers(
    lang_code,
    sentences,
    model,
    stride,
    block_size,
    batch_size,
    k: int = 3,
    sample_pct: float = 0.1,
    max_n_samples: int = sys.maxsize,
    use_pysbd=False,
    positive_index=None,
    do_lowercase=False,
    do_remove_punct=False,
    threshold: float = 0.01,
):
    if positive_index is None:
        positive_index = Constants.NEWLINE_INDEX

    # Preprocess sentences
    sentences = [sentence.lstrip("-").strip() for sentence in sentences]

    separator = Constants.SEPARATORS[lang_code]
    metrics_list = []
    accuracy_list = []
    accuracy_list_optimal = []
    info_list = []

    # get pairs of sentences (non-overlapping)
    sampled_k_mers = generate_k_mers(
        sentences=sentences,
        k=k,
        do_lowercase=do_lowercase,
        do_remove_punct=do_remove_punct,
        sample_pct=sample_pct,
        max_n_samples=max_n_samples,
        min_k_mer_length=0,
    )

    # get logits for each pair
    logits, n_tokens_list = process_logits_k_mers(
        pairs=sampled_k_mers,
        model=PyTorchWrapper(model.backbone),
        lang_code=lang_code,
        block_size=block_size,
        batch_size=batch_size,
        verbose=False,
    )

    for i, k_mer in enumerate(sampled_k_mers):
        newline_probs = logits[i][:, positive_index]

        k_mer_text = separator.join(k_mer)
        true_end_indices = np.cumsum(np.array([len(s) for s in k_mer])) + np.arange(len(k_mer)) * len(separator)
        newline_labels = np.zeros(len(k_mer_text))
        newline_labels[true_end_indices - 1] = 1

        # Get metrics for the k-mer
        k_mer_metrics, info = get_metrics(newline_labels, newline_probs, threshold=threshold)
        metrics_list.append(k_mer_metrics["pr_auc"])
        info_list.append(info)

        predicted_labels = newline_probs > np.log(threshold / (1 - threshold))  # inverse sigmoid
        predicted_labels_optimal = newline_probs > np.log(
            info["threshold_best"] / (1 - info["threshold_best"])
        )  # inverse sigmoid
        # For accuracy, check if all the labels in between are correctly predicted (ignore the one at the end)
        intermediate_newline_labels = newline_labels[:-1]  # Exclude the end
        intermediate_predicted_labels = predicted_labels[:-1]
        intermediate_predicted_labels_opt = predicted_labels_optimal[:-1]
        correct = np.array_equal(intermediate_newline_labels, intermediate_predicted_labels)
        correct_optimal = np.array_equal(intermediate_newline_labels, intermediate_predicted_labels_opt)
        accuracy_list.append(correct)
        accuracy_list_optimal.append(correct_optimal)

    # Compute and return the average metric and accuracy
    average_metric = np.mean(metrics_list)
    avg_accuracy = np.mean(accuracy_list)
    # get averages for info_list
    # print(len(info_list), len(sampled_k_mers))
    if len(sampled_k_mers) > 0:
        avg_info = {
            key: np.mean([info[key] for info in info_list])
            for key in info_list[0].keys()
            if isinstance(info_list[0][key], (int, float))
        }
        avg_info["accuracy_optimal"] = np.mean(accuracy_list_optimal)
    else:
        avg_info = {}
        avg_info["f1"] = 0
        avg_info["f1_best"] = 0
        avg_info["threshold_best"] = 0
    return average_metric, avg_accuracy, avg_info



import os
from transformers import AutoTokenizer
from tokenizers import AddedToken
from wtpsplit.utils import Constants, LabelArgs


def get_subword_label_dict(label_args, tokenizer):
    label_dict = {}

    n_unks = 0
    # Map auxiliary characters to token IDs with labels
    for i, c in enumerate(Constants.PUNCTUATION_CHARS):
        token_id = tokenizer.convert_tokens_to_ids(c)
        label_dict[token_id] = 1 + Constants.AUX_OFFSET + i

        print(
            f"auxiliary character {c} has token ID {token_id} and label {label_dict[token_id]}, decoded: {tokenizer.decode([token_id])}"
        )
        if token_id == tokenizer.unk_token_id:
            n_unks += 1

    print(f"found {n_unks} UNK tokens in auxiliary characters")

    # Map newline characters to token IDs with labels
    for c in label_args.newline_chars:
        token_id = tokenizer.convert_tokens_to_ids(c)
        label_dict[token_id] = 1 + Constants.NEWLINE_INDEX
        print(f"newline character {c} has token ID {token_id} and label {label_dict[token_id]}, decoded:")
        print(r"{}".format(tokenizer.decode([token_id])))

    return label_dict


tokenizer = AutoTokenizer.from_pretrained("xlm-roberta-base")
tokenizer.add_special_tokens({"additional_special_tokens": [AddedToken("\n")]})

label_dict = get_subword_label_dict(LabelArgs(), tokenizer)
print(len(label_dict))


def write_punctuation_file():
    with open(os.path.join(Constants.ROOT_DIR, "punctuation_xlmr.txt"), "w", encoding="utf-8") as file:
        for char in Constants.PUNCTUATION_CHARS:
            token_id = tokenizer.convert_tokens_to_ids(char)
            if token_id != tokenizer.unk_token_id:
                file.write(char + "\n")


def write_punctuation_file_unk():
    added_unk = False
    with open(os.path.join(Constants.ROOT_DIR, "punctuation_xlmr_unk.txt"), "w", encoding="utf-8") as file:
        for char in Constants.PUNCTUATION_CHARS:
            token_id = tokenizer.convert_tokens_to_ids(char)
            if token_id != tokenizer.unk_token_id:
                file.write(char + "\n")
            elif not added_unk:
                print("added unk")
                file.write("<unk>\n")
                added_unk = True


write_punctuation_file()
write_punctuation_file_unk()

label_args_default = LabelArgs()
print(Constants.PUNCTUATION_CHARS, len(Constants.PUNCTUATION_CHARS))

label_args_custom = LabelArgs(custom_punctuation_file="punctuation_xlmr.txt")
print(Constants.PUNCTUATION_CHARS, len(Constants.PUNCTUATION_CHARS))



import contextlib
import math
import os
import warnings
from pathlib import Path

# avoid the "None of PyTorch, TensorFlow, etc. have been found" warning.
with contextlib.redirect_stderr(open(os.devnull, "w")):
    import transformers  # noqa

import numpy as np
import skops.io as sio

from huggingface_hub import hf_hub_download
from transformers import AutoConfig, AutoModelForTokenClassification, AutoTokenizer
from transformers.utils.hub import cached_file

from wtpsplit.extract import BertCharORTWrapper, SaTORTWrapper, PyTorchWrapper, extract
from wtpsplit.utils import Constants, indices_to_sentences, sigmoid, token_to_char_probs

__version__ = "2.1.2"

warnings.simplefilter("default", DeprecationWarning)  # show by default
warnings.simplefilter("ignore", category=FutureWarning)  # for tranformers


class WtP:
    def __init__(
        self,
        model_name_or_model,
        from_pretrained_kwargs=None,
        ort_providers=None,
        ort_kwargs=None,
        mixtures=None,
        hub_prefix="benjamin",
        ignore_legacy_warning=False,
    ):
        self.model_name_or_model = model_name_or_model
        self.ort_providers = ort_providers
        self.ort_kwargs = ort_kwargs

        mixture_path = None

        if not ignore_legacy_warning:
            # WtP is deprecated!
            warnings.warn(
                "You are using WtP, the old sentence segmentation model. "
                "It is highly encouraged to use SaT instead due to strongly improved performance and efficiency. "
                "See https://github.com/segment-any-text/wtpsplit for more info. "
                "To ignore this warning, set ignore_legacy_warning=True.",
                DeprecationWarning,
            )

        if isinstance(model_name_or_model, (str, Path)):
            model_name = str(model_name_or_model)
            is_local = os.path.isdir(model_name)

            if not is_local and hub_prefix is not None:
                model_name_to_fetch = f"{hub_prefix}/{model_name}"
            else:
                model_name_to_fetch = model_name

            if is_local:
                model_path = Path(model_name)
                mixture_path = model_path / "mixtures.skops"
                if not mixture_path.exists():
                    mixture_path = None
                onnx_path = model_path / "model.onnx"
                if not onnx_path.exists():
                    onnx_path = None
            else:
                try:
                    mixture_path = cached_file(model_name_to_fetch, "mixtures.skops", **(from_pretrained_kwargs or {}))
                except OSError:
                    mixture_path = None

                # no need to load if no ort_providers set
                if ort_providers is not None:
                    onnx_path = cached_file(model_name_to_fetch, "model.onnx", **(from_pretrained_kwargs or {}))
                else:
                    onnx_path = None

            if ort_providers is not None:
                if onnx_path is None:
                    raise ValueError(
                        "Could not find an ONNX model in the model directory. Try `use_ort=False` to run with PyTorch."
                    )

                try:
                    import onnxruntime as ort  # noqa
                except ModuleNotFoundError:
                    raise ValueError("Please install `onnxruntime` to use WtP with an ONNX model.")

                # to register models for AutoConfig
                import wtpsplit.configs  # noqa

                self.model = BertCharORTWrapper(
                    AutoConfig.from_pretrained(model_name_to_fetch, **(from_pretrained_kwargs or {})),
                    ort.InferenceSession(str(onnx_path), providers=ort_providers, **(ort_kwargs or {})),
                )
            else:
                # to register models for AutoConfig
                try:
                    import torch  # noqa
                except ModuleNotFoundError:
                    raise ValueError("Please install `torch` to use WtP with a PyTorch model.")

                import wtpsplit.models  # noqa

                self.model = PyTorchWrapper(
                    AutoModelForTokenClassification.from_pretrained(
                        model_name_to_fetch, **(from_pretrained_kwargs or {})
                    )
                )
        else:
            if ort_providers is not None:
                raise ValueError("You can only use onnxruntime with a model directory, not a model object.")

            self.model = model_name_or_model

        if mixtures is not None:
            self.mixtures = mixtures
        elif mixture_path is not None:
            self.mixtures = sio.load(
                mixture_path,
                ["numpy.float32", "numpy.float64", "sklearn.linear_model._logistic.LogisticRegression"],
            )
        else:
            self.mixtures = None

    def __getattr__(self, name):
        assert hasattr(self, "model")
        return getattr(self.model, name)

    def predict_proba(
        self,
        text_or_texts,
        lang_code: str = None,
        style: str = None,
        stride=256,
        block_size: int = 512,
        batch_size=32,
        pad_last_batch: bool = False,
        remove_whitespace_before_inference: bool = False,
        outer_batch_size=1000,
        return_paragraph_probabilities=False,
        verbose: bool = False,
    ):
        if isinstance(text_or_texts, str):
            return next(
                self._predict_proba(
                    [text_or_texts],
                    lang_code=lang_code,
                    style=style,
                    stride=stride,
                    block_size=block_size,
                    batch_size=batch_size,
                    pad_last_batch=pad_last_batch,
                    remove_whitespace_before_inference=remove_whitespace_before_inference,
                    outer_batch_size=outer_batch_size,
                    return_paragraph_probabilities=return_paragraph_probabilities,
                    verbose=verbose,
                )
            )
        else:
            return self._predict_proba(
                text_or_texts,
                lang_code=lang_code,
                style=style,
                stride=stride,
                block_size=block_size,
                batch_size=batch_size,
                pad_last_batch=pad_last_batch,
                remove_whitespace_before_inference=remove_whitespace_before_inference,
                outer_batch_size=outer_batch_size,
                return_paragraph_probabilities=return_paragraph_probabilities,
                verbose=verbose,
            )

    def _predict_proba(
        self,
        texts,
        lang_code: str,
        style: str,
        stride: int,
        block_size: int,
        batch_size: int,
        pad_last_batch: bool,
        remove_whitespace_before_inference: bool,
        outer_batch_size: int,
        return_paragraph_probabilities: bool,
        verbose: bool,
    ):
        if style is not None:
            if lang_code is None:
                raise ValueError("Please specify a `lang_code` when passing a `style` to adapt to.")

            if self.mixtures is None:
                raise ValueError(
                    "This model does not have any associated mixtures. Maybe they are missing from the model directory?"
                )

            try:
                clf, _, _, _ = self.mixtures[lang_code][style]
            except KeyError:
                raise ValueError(f"Could not find a mixture for the style '{style}'.")
        else:
            clf = None

        n_outer_batches = math.ceil(len(texts) / outer_batch_size)

        for outer_batch_idx in range(n_outer_batches):
            start, end = outer_batch_idx * outer_batch_size, min((outer_batch_idx + 1) * outer_batch_size, len(texts))

            outer_batch_texts = texts[start:end]
            input_texts = []
            space_positions = []

            for text in outer_batch_texts:
                if remove_whitespace_before_inference:
                    text_space_positions = []
                    input_text = ""

                    for c in text:
                        if c == " ":
                            text_space_positions.append(len(input_text) + len(text_space_positions))
                        else:
                            input_text += c

                    space_positions.append(text_space_positions)
                else:
                    input_text = text

                input_texts.append(input_text)

            empty_string_indices = [i for i, text in enumerate(input_texts) if not text.strip()]
            # remove empty strings from input_texts
            input_texts = [text for text in input_texts if text.strip()]

            if input_texts:
                outer_batch_logits = extract(
                    input_texts,
                    self.model,
                    lang_code=lang_code,
                    stride=stride,
                    max_block_size=block_size,
                    batch_size=batch_size,
                    pad_last_batch=pad_last_batch,
                    verbose=verbose,
                )[0]
            else:
                outer_batch_logits = []

            def newline_probability_fn(logits):
                return sigmoid(logits[:, Constants.NEWLINE_INDEX])

            # add back empty strings
            for i in empty_string_indices:
                outer_batch_logits.insert(i, np.ones([1, 1]) * -np.inf)

            for i, (text, logits) in enumerate(zip(outer_batch_texts, outer_batch_logits)):
                if style is not None:
                    sentence_probs = clf.predict_proba(logits)[:, 1]
                    newline_probs = newline_probability_fn(logits)
                else:
                    sentence_probs = newline_probs = newline_probability_fn(logits)

                if remove_whitespace_before_inference:
                    full_newline_probs, full_sentence_probs = list(newline_probs), list(sentence_probs)

                    for j in space_positions[i]:
                        full_newline_probs.insert(j, np.zeros_like(newline_probs[0]))
                        full_sentence_probs.insert(j, np.zeros_like(sentence_probs[0]))

                    newline_probs = np.array(full_newline_probs)
                    sentence_probs = np.array(full_sentence_probs)

                if return_paragraph_probabilities:
                    yield sentence_probs, newline_probs
                else:
                    yield sentence_probs

    def split(
        self,
        text_or_texts,
        lang_code: str = None,
        style: str = None,
        threshold: float = None,
        stride=64,
        block_size: int = 512,
        batch_size=32,
        pad_last_batch: bool = False,
        remove_whitespace_before_inference: bool = False,
        outer_batch_size=1000,
        paragraph_threshold: float = 0.5,
        strip_whitespace: bool = False,
        do_paragraph_segmentation=False,
        verbose: bool = False,
    ):
        if isinstance(text_or_texts, str):
            return next(
                self._split(
                    [text_or_texts],
                    lang_code=lang_code,
                    style=style,
                    threshold=threshold,
                    stride=stride,
                    block_size=block_size,
                    batch_size=batch_size,
                    pad_last_batch=pad_last_batch,
                    remove_whitespace_before_inference=remove_whitespace_before_inference,
                    outer_batch_size=outer_batch_size,
                    paragraph_threshold=paragraph_threshold,
                    strip_whitespace=strip_whitespace,
                    do_paragraph_segmentation=do_paragraph_segmentation,
                    verbose=verbose,
                )
            )
        else:
            return self._split(
                text_or_texts,
                lang_code=lang_code,
                style=style,
                threshold=threshold,
                stride=stride,
                block_size=block_size,
                batch_size=batch_size,
                pad_last_batch=pad_last_batch,
                remove_whitespace_before_inference=remove_whitespace_before_inference,
                outer_batch_size=outer_batch_size,
                paragraph_threshold=paragraph_threshold,
                strip_whitespace=strip_whitespace,
                do_paragraph_segmentation=do_paragraph_segmentation,
                verbose=verbose,
            )

    def get_threshold(self, lang_code: str, style: str, return_punctuation_threshold: bool = False):
        try:
            _, _, punctuation_threshold, threshold = self.mixtures[lang_code][style]
        except KeyError:
            raise ValueError(f"Could not find a mixture for the style '{style}' and language '{lang_code}'.")

        if return_punctuation_threshold:
            return punctuation_threshold

        return threshold

    def _split(
        self,
        texts,
        lang_code: str,
        style: str,
        threshold: float,
        stride: int,
        block_size: int,
        batch_size: int,
        pad_last_batch: bool,
        remove_whitespace_before_inference: bool,
        outer_batch_size: int,
        paragraph_threshold: float,
        do_paragraph_segmentation: bool,
        strip_whitespace: bool,
        verbose: bool,
    ):
        if style is not None:
            if lang_code is None:
                raise ValueError("Please specify a `lang_code` when passing a `style` to adapt to.")

            if self.mixtures is None:
                raise ValueError(
                    "This model does not have any associated mixtures. Maybe they are missing from the model directory?"
                )

            try:
                _, _, default_threshold, _ = self.mixtures[lang_code][style]
            except KeyError:
                raise ValueError(f"Could not find a mixture for the style '{style}'.")
        else:
            # the established default for newline prob threshold is 0.01
            default_threshold = 0.01

        sentence_threshold = threshold if threshold is not None else default_threshold

        for text, probs in zip(
            texts,
            self.predict_proba(
                texts,
                lang_code=lang_code,
                style=style,
                stride=stride,
                block_size=block_size,
                batch_size=batch_size,
                pad_last_batch=pad_last_batch,
                remove_whitespace_before_inference=remove_whitespace_before_inference,
                outer_batch_size=outer_batch_size,
                return_paragraph_probabilities=do_paragraph_segmentation,
                verbose=verbose,
            ),
        ):
            if do_paragraph_segmentation:
                sentence_probs, newline_probs = probs

                offset = 0

                paragraphs = []

                for paragraph in indices_to_sentences(text, np.where(newline_probs > paragraph_threshold)[0]):
                    sentences = []

                    for sentence in indices_to_sentences(
                        paragraph,
                        np.where(
                            sentence_probs[offset : offset + len(paragraph)] > sentence_threshold,
                        )[0],
                        strip_whitespace=strip_whitespace,
                    ):
                        sentences.append(sentence)

                    paragraphs.append(sentences)
                    offset += len(paragraph)

                yield paragraphs
            else:
                sentences = indices_to_sentences(
                    text, np.where(probs > sentence_threshold)[0], strip_whitespace=strip_whitespace
                )
                yield sentences


class SaT:
    def __init__(
        self,
        model_name_or_model,
        from_pretrained_kwargs=None,
        ort_providers=None,
        ort_kwargs=None,
        style_or_domain: str = None,
        language: str = None,
        lora_path: str = None,  # local
        hub_prefix="segment-any-text",
    ):
        self.model_name_or_model = model_name_or_model
        self.ort_providers = ort_providers
        self.ort_kwargs = ort_kwargs

        self.use_lora = False

        self.tokenizer = AutoTokenizer.from_pretrained("facebookAI/xlm-roberta-base")

        if isinstance(model_name_or_model, (str, Path)):
            model_name = str(model_name_or_model)
            is_local = os.path.isdir(model_name)

            if not is_local and hub_prefix is not None:
                model_name_to_fetch = f"{hub_prefix}/{model_name}"
            else:
                model_name_to_fetch = model_name

            if is_local:
                model_path = Path(model_name)
                onnx_path = model_path / "model_optimized.onnx"
                if not onnx_path.exists():
                    onnx_path = None
            else:
                # no need to load if no ort_providers set
                if ort_providers is not None:
                    onnx_path = cached_file(
                        model_name_to_fetch, "model_optimized.onnx", **(from_pretrained_kwargs or {})
                    )
                else:
                    onnx_path = None

            if ort_providers is not None:
                if onnx_path is None:
                    raise ValueError(
                        "Could not find an ONNX model in the model directory. Try `use_ort=False` to run with PyTorch."
                    )

                try:
                    import onnxruntime as ort  # noqa
                except ModuleNotFoundError:
                    raise ValueError("Please install `onnxruntime` to use SaT with an ONNX model.")

                # to register models for AutoConfig
                import wtpsplit.configs  # noqa

                self.model = SaTORTWrapper(
                    AutoConfig.from_pretrained(model_name_to_fetch, **(from_pretrained_kwargs or {})),
                    ort.InferenceSession(str(onnx_path), providers=ort_providers, **(ort_kwargs or {})),
                )
                if lora_path:
                    raise ValueError(
                        "If using ONNX with LoRA, execute `scripts/export_to_onnx_sat.py` with `use_lora=True`."
                        "Reference the chosen `output_dir` here for `model_name_or_model`. and set `lora_path=None`."
                    )
            else:
                # to register models for AutoConfig
                try:
                    import torch  # noqa
                except ModuleNotFoundError:
                    raise ValueError("Please install `torch` to use WtP with a PyTorch model.")

                import wtpsplit.models  # noqa

                self.model = PyTorchWrapper(
                    AutoModelForTokenClassification.from_pretrained(
                        model_name_to_fetch, **(from_pretrained_kwargs or {})
                    )
                )
            # LoRA LOADING
            if not lora_path:
                if (style_or_domain and not language) or (language and not style_or_domain):
                    raise ValueError("Please specify both language and style_or_domain!")
            if (style_or_domain and language) or lora_path:
                import adapters  # noqa
                from adapters.models import MODEL_MIXIN_MAPPING  # noqa
                from adapters.models.bert.mixin_bert import BertModelAdaptersMixin  # noqa

                # monkey patch mixin to avoid forking whole adapters library
                MODEL_MIXIN_MAPPING["SubwordXLMRobertaModel"] = BertModelAdaptersMixin
                model_type = self.model.model.config.model_type
                # adapters need xlm-roberta as model type.
                self.model.model.config.model_type = "xlm-roberta"
                adapters.init(self.model.model)
                # reset model type (used later)
                self.model.model.config.model_type = model_type
                try:
                    if not lora_path:
                        for file in [
                            "adapter_config.json",
                            "head_config.json",
                            "pytorch_adapter.bin",
                            "pytorch_model_head.bin",
                        ]:
                            hf_hub_download(
                                repo_id=model_name_to_fetch,
                                subfolder=f"loras/{style_or_domain}/{language}",
                                filename=file,
                                local_dir=Constants.CACHE_DIR,
                            )
                        lora_load_path = str(Constants.CACHE_DIR / "loras" / style_or_domain / language)
                    else:
                        lora_load_path = lora_path

                    self.model.model.load_adapter(
                        lora_load_path,
                        set_active=True,
                        with_head=True,
                        load_as="sat-lora",
                    )
                    # merge lora weights into transformer for 0 efficiency overhead
                    self.model.model.merge_adapter("sat-lora")
                    self.use_lora = True
                except:  # noqa
                    if lora_path:
                        print(f"LoRA at {lora_path} not found, using base model...")
                    else:
                        print(f"LoRA {style_or_domain}/{language} not found, using base model...")
        else:
            if ort_providers is not None:
                raise ValueError("You can only use onnxruntime with a model directory, not a model object.")

            self.model = model_name_or_model

    def __getattr__(self, name):
        assert hasattr(self, "model")
        return getattr(self.model, name)

    def predict_proba(
        self,
        text_or_texts,
        stride=256,
        block_size: int = 512,
        batch_size=32,
        pad_last_batch: bool = False,
        remove_whitespace_before_inference: bool = False,
        outer_batch_size=1000,
        return_paragraph_probabilities=False,
        verbose: bool = False,
    ):
        if isinstance(text_or_texts, str):
            return next(
                self._predict_proba(
                    [text_or_texts],
                    stride=stride,
                    block_size=block_size,
                    batch_size=batch_size,
                    pad_last_batch=pad_last_batch,
                    remove_whitespace_before_inference=remove_whitespace_before_inference,
                    outer_batch_size=outer_batch_size,
                    return_paragraph_probabilities=return_paragraph_probabilities,
                    verbose=verbose,
                )
            )
        else:
            return self._predict_proba(
                text_or_texts,
                stride=stride,
                block_size=block_size,
                batch_size=batch_size,
                pad_last_batch=pad_last_batch,
                remove_whitespace_before_inference=remove_whitespace_before_inference,
                outer_batch_size=outer_batch_size,
                return_paragraph_probabilities=return_paragraph_probabilities,
                verbose=verbose,
            )

    def _predict_proba(
        self,
        texts,
        stride: int,
        block_size: int,
        batch_size: int,
        pad_last_batch: bool,
        remove_whitespace_before_inference: bool,
        outer_batch_size: int,
        return_paragraph_probabilities: bool,
        verbose: bool,
    ):
        def newline_probability_fn(logits):
            return sigmoid(logits[:, Constants.NEWLINE_INDEX])

        n_outer_batches = math.ceil(len(texts) / outer_batch_size)

        for outer_batch_idx in range(n_outer_batches):
            start, end = outer_batch_idx * outer_batch_size, min((outer_batch_idx + 1) * outer_batch_size, len(texts))

            outer_batch_texts = texts[start:end]
            input_texts = []
            space_positions = []

            for text in outer_batch_texts:
                if remove_whitespace_before_inference:
                    text_space_positions = []
                    input_text = ""

                    for c in text:
                        if c == " ":
                            text_space_positions.append(len(input_text) + len(text_space_positions))
                        else:
                            input_text += c

                    space_positions.append(text_space_positions)
                else:
                    input_text = text

                input_texts.append(input_text)

            empty_string_indices = [i for i, text in enumerate(input_texts) if not text.strip()]
            # remove empty strings from input_texts
            input_texts = [text for text in input_texts if text.strip()]
            if input_texts:
                outer_batch_logits, _, tokenizer, tokenizer_output = extract(
                    input_texts,
                    self.model,
                    stride=stride,
                    max_block_size=block_size,
                    batch_size=batch_size,
                    pad_last_batch=pad_last_batch,
                    verbose=verbose,
                    tokenizer=self.tokenizer,
                )

                # convert token probabilities to character probabilities for the entire array
                outer_batch_logits = [
                    token_to_char_probs(
                        input_texts[i],
                        tokenizer_output["input_ids"][i],
                        outer_batch_logits[i],
                        tokenizer,
                        tokenizer_output["offset_mapping"][i],
                    )
                    for i in range(len(input_texts))
                ]
            else:
                outer_batch_logits = []

            # add back empty strings
            for i in empty_string_indices:
                outer_batch_logits.insert(i, np.ones([1, 1]) * -np.inf)

            for i, (text, logits) in enumerate(zip(outer_batch_texts, outer_batch_logits)):
                sentence_probs = newline_probs = newline_probability_fn(logits)

                if remove_whitespace_before_inference:
                    full_newline_probs, full_sentence_probs = list(newline_probs), list(sentence_probs)

                    for j in space_positions[i]:
                        full_newline_probs.insert(j, np.zeros_like(newline_probs[0]))
                        full_sentence_probs.insert(j, np.zeros_like(sentence_probs[0]))

                    newline_probs = np.array(full_newline_probs)
                    sentence_probs = np.array(full_sentence_probs)

                if return_paragraph_probabilities:
                    yield sentence_probs, newline_probs
                else:
                    yield sentence_probs

    def split(
        self,
        text_or_texts,
        threshold: float = None,
        stride=64,
        block_size: int = 512,
        batch_size=32,
        pad_last_batch: bool = False,
        remove_whitespace_before_inference: bool = False,
        outer_batch_size=1000,
        paragraph_threshold: float = 0.5,
        strip_whitespace: bool = False,
        do_paragraph_segmentation: bool = False,
        treat_newline_as_space: bool = False,
        verbose: bool = False,
    ):
        if isinstance(text_or_texts, str):
            return next(
                self._split(
                    [text_or_texts],
                    threshold=threshold,
                    stride=stride,
                    block_size=block_size,
                    batch_size=batch_size,
                    pad_last_batch=pad_last_batch,
                    remove_whitespace_before_inference=remove_whitespace_before_inference,
                    outer_batch_size=outer_batch_size,
                    paragraph_threshold=paragraph_threshold,
                    strip_whitespace=strip_whitespace,
                    do_paragraph_segmentation=do_paragraph_segmentation,
                    treat_newline_as_space=treat_newline_as_space,
                    verbose=verbose,
                )
            )
        else:
            return self._split(
                text_or_texts,
                threshold=threshold,
                stride=stride,
                block_size=block_size,
                batch_size=batch_size,
                pad_last_batch=pad_last_batch,
                remove_whitespace_before_inference=remove_whitespace_before_inference,
                outer_batch_size=outer_batch_size,
                paragraph_threshold=paragraph_threshold,
                strip_whitespace=strip_whitespace,
                do_paragraph_segmentation=do_paragraph_segmentation,
                treat_newline_as_space=treat_newline_as_space,
                verbose=verbose,
            )

    def _split(
        self,
        texts,
        threshold: float,
        stride: int,
        block_size: int,
        batch_size: int,
        pad_last_batch: bool,
        paragraph_threshold: float,
        remove_whitespace_before_inference: bool,
        outer_batch_size: int,
        do_paragraph_segmentation: bool,
        treat_newline_as_space: bool,
        strip_whitespace: bool,
        verbose: bool,
    ):
        def get_default_threshold(model_str: str):
            if "sm" in model_str:
                return 0.25
            if self.use_lora:
                return 0.5
            if "no-limited-lookahead" in model_str and "sm" not in model_str:
                return 0.01
            return 0.025

        default_threshold = get_default_threshold(self.model_name_or_model)
        sentence_threshold = threshold if threshold is not None else default_threshold

        for text, probs in zip(
            texts,
            self.predict_proba(
                texts,
                stride=stride,
                block_size=block_size,
                batch_size=batch_size,
                pad_last_batch=pad_last_batch,
                remove_whitespace_before_inference=remove_whitespace_before_inference,
                outer_batch_size=outer_batch_size,
                return_paragraph_probabilities=do_paragraph_segmentation,
                verbose=verbose,
            ),
        ):
            if do_paragraph_segmentation:
                sentence_probs, newline_probs = probs

                offset = 0
                paragraphs = []

                for paragraph in indices_to_sentences(text, np.where(newline_probs > paragraph_threshold)[0]):
                    sentences = []

                    for sentence in indices_to_sentences(
                        paragraph,
                        np.where(
                            sentence_probs[offset : offset + len(paragraph)] > sentence_threshold,
                        )[0],
                        strip_whitespace=strip_whitespace,
                    ):
                        sentences.append(sentence)

                    paragraphs.append(sentences)
                    offset += len(paragraph)

                yield paragraphs
            else:
                sentences = indices_to_sentences(
                    text, np.where(probs > sentence_threshold)[0], strip_whitespace=strip_whitespace
                )
                if not treat_newline_as_space:
                    # within the model, newlines in the text were ignored - they were treated as spaces.
                    # this is the default behavior: additionally split on newlines as provided in the input
                    new_sentences = []
                    for sentence in sentences:
                        new_sentences.extend(sentence.split("\n"))
                    sentences = new_sentences
                else:
                    warnings.warn(
                        "treat_newline_as_space=True will lead to newlines in the output "
                        "if they were present in the input. Within the model, such newlines are "
                        "treated as spaces. "
                        "If you want to split on such newlines, set treat_newline_as_space=False."
                    )
                yield sentences



from setuptools import setup, find_packages

setup(
    name="wtpsplit",
    version="2.1.2",
    packages=find_packages(),
    description="Universal Robust, Efficient and Adaptable Sentence Segmentation",
    author="Markus Frohmann, Igor Sterner, Benjamin Minixhofer",
    author_email="markus.frohmann@gmail.com",
    install_requires=[
        "onnxruntime>=1.13.1",
        "transformers>=4.22.2",
        "huggingface-hub==0.25.2",   # see https://github.com/segment-any-text/wtpsplit/issues/135
        "numpy>=1.0",
        "scikit-learn>=1",
        "tqdm",
        "skops",
        "pandas>=1",
        "cached_property",  # for Py37
        "mosestokenizer",
        "adapters",
    ],
    url="https://github.com/segment-any-text/wtpsplit",
    package_data={"wtpsplit": ["data/*"]},
    include_package_data=True,
    license="MIT",
)



import json
import logging
import os
import random
from collections import defaultdict
from dataclasses import dataclass, field
from pathlib import Path
from typing import List

import numpy as np
import pandas as pd
from cached_property import cached_property
from mosestokenizer import MosesTokenizer
from transformers import AutoTokenizer

# same as in CANINE
PRIMES = [31, 43, 59, 61, 73, 97, 103, 113, 137, 149, 157, 173, 181, 193, 211, 223]

logger = logging.getLogger(__name__)


class ConstantsClass:
    NEWLINE_INDEX = 0
    AUX_OFFSET = 1
    DEFAULT_PUNCTUATION_FILE = "punctuation.txt"
    _PUNCTUATION_FILE = "punctuation.txt"

    @classmethod
    def set_punctuation_file(cls, file_name):
        cls._PUNCTUATION_FILE = file_name

    @cached_property
    def ROOT_DIR(self):
        return Path(os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))

    @cached_property
    def CACHE_DIR(self):
        CACHE_DIR = self.ROOT_DIR / ".cache"
        CACHE_DIR.mkdir(exist_ok=True)
        return CACHE_DIR

    @cached_property
    def LANGINFO(self):
        return pd.read_csv(os.path.join(self.ROOT_DIR, "data", "language_info.csv"), index_col=0)

    @property
    def PUNCTUATION_CHARS(self):
        punctuation_path = os.path.join(self.ROOT_DIR, "data", self._PUNCTUATION_FILE)
        if os.path.exists(punctuation_path):
            return [x.strip() for x in open(punctuation_path).readlines()]
        else:
            raise FileNotFoundError(f"The file {punctuation_path} does not exist.")

    @cached_property
    def PUNCTUATION_MAP(self):
        return json.load(open(os.path.join(self.ROOT_DIR, "data", "punctuation.json")))

    @cached_property
    def LANG_CODE_TO_INDEX(self):
        return {lang: i for i, lang in enumerate(self.LANGINFO.index)}

    @cached_property
    def SEPARATORS(self):
        return defaultdict(
            lambda: " ", {lang: ("" if row["no_whitespace"] else " ") for lang, row in self.LANGINFO.iterrows()}
        )


Constants = ConstantsClass()


@dataclass
class LabelArgs:
    newline_remove_prob: float = 1.0
    auxiliary_remove_prob: float = 0.5
    hyphen_remove_prob: float = 0.9
    newline_whitespace_prob: float = 0.99
    hyphen_smooth_prob: float = 0.9
    newline_chars: List[str] = field(default_factory=lambda: ["\n"])
    auxiliary_chars: List[str] = field(default_factory=lambda: [])
    hyphen_chars: List[str] = field(default_factory=lambda: ["-", "‐"])
    use_auxiliary: bool = False
    custom_punctuation_file: str = None
    retain_first_consecutive_punctuation: bool = True
    non_whitespace_remove_spaces: bool = True
    corrupt_entire_chunk_prob: float = 0.0
    corrupt_entire_chunk_strategy: str = "mix"
    corrupt_entire_chunk_prob_full: float = 0.5

    def __post_init__(self):
        if self.custom_punctuation_file:
            Constants.set_punctuation_file(self.custom_punctuation_file)
        else:
            Constants.set_punctuation_file("punctuation.txt")
        self.auxiliary_chars = Constants.DEFAULT_PUNCTUATION_FILE


def get_label_dict(label_args):
    label_dict = {}

    for i, c in enumerate(Constants.PUNCTUATION_CHARS):
        label_dict[ord(c)] = 1 + Constants.AUX_OFFSET + i

    for c in label_args.newline_chars:
        label_dict[ord(c)] = 1 + Constants.NEWLINE_INDEX

    return label_dict


def get_subword_label_dict(label_args, tokenizer):
    label_dict = {}

    n_unks = 0
    # Map auxiliary characters to token IDs with labels
    logger.info(f"Using {Constants.PUNCTUATION_CHARS} auxiliary characters.")
    for i, c in enumerate(Constants.PUNCTUATION_CHARS):
        token_id = tokenizer.convert_tokens_to_ids(c)
        label_dict[token_id] = 1 + Constants.AUX_OFFSET + i

        if token_id == tokenizer.unk_token_id:
            n_unks += 1

    logger.info(f"found {n_unks} UNK tokens in auxiliary characters")

    # Map newline characters to token IDs with labels
    for c in label_args.newline_chars:
        token_id = tokenizer.convert_tokens_to_ids(c)
        label_dict[token_id] = 1 + Constants.NEWLINE_INDEX
        logger.info(f"newline character {c} has token ID {token_id} and label {label_dict[token_id]}, decoded:")
        logger.info(f"{tokenizer.decode([token_id])}")

    return label_dict


def sigmoid(x):
    return 1 / (1 + np.exp(-x.astype(np.float32)))  # fp32 for better precision


def encode(text):
    return [ord(c) for c in text]


def hash_encode(encoding, num_hashes=8, num_buckets=8192):
    if num_hashes > len(PRIMES):
        raise ValueError(f"`num_hashes` must be <= {len(PRIMES)}")

    hash_ids = np.zeros((len(encoding), num_hashes), dtype=np.int64)
    for i in range(num_hashes):
        shard_ids = (encoding + 1) * PRIMES[i]
        hash_ids[:, i] = shard_ids % num_buckets

    return hash_ids


def label(input_ids, label_dict):
    return [label_dict.get(input_id, 0) for input_id in input_ids[1:]] + [0]


def lang_code_to_lang(lang_code):
    from iso639 import languages

    if lang_code == "el":
        return "Greek"
    elif lang_code == "tl":
        return "Filipino"
    elif lang_code == "ug":
        return "Uyghur"

    try:
        return languages.get(alpha2=lang_code).name
    except KeyError:
        return languages.get(part3=lang_code).name


def corrupt(text: str, do_lowercase: bool, do_remove_punct: bool):
    if do_lowercase:
        text = text.lower()
    if do_remove_punct:
        for punct in Constants.PUNCTUATION_CHARS:
            text = text.replace(punct, "")
    return text


def corrupt_asr(text: str, lang):
    if text is None:
        return None

    separator = Constants.SEPARATORS.get(lang, " ")

    sentences = text.split("\n")

    if separator == "":
        corrupted_sentences = [
            "".join([char for char in sentence if char not in Constants.PUNCTUATION_CHARS]).lower()
            for sentence in sentences
        ]
        return corrupted_sentences

    try:
        tokenizer = MosesTokenizer(lang)

        tokenized_sentences = [tokenizer.tokenize(sentence) for sentence in sentences]
        corrupted_tokenized_sentences = [
            [token for token in tokens if token not in Constants.PUNCTUATION_CHARS] for tokens in tokenized_sentences
        ]
        corrupted_sentences = [
            tokenizer.detokenize(corrupted_tokens).lower() for corrupted_tokens in corrupted_tokenized_sentences
        ]
    except:  # noqa
        corrupted_sentences = [
            "".join([char for char in sentence if char not in Constants.PUNCTUATION_CHARS]).lower()
            for sentence in sentences
        ]
        return corrupted_sentences

    return corrupted_sentences


def corrupt_training(
    input_ids,
    block_ids,
    lang,
    label_args,
    label_dict,
    pack_samples=False,
    min_length=None,
    tokenizer=None,
):
    input_ids = input_ids.copy()
    block_ids = block_ids.copy()
    if random.random() < label_args.corrupt_entire_chunk_prob:
        # choose corruption strategy
        if label_args.corrupt_entire_chunk_strategy == "mix":
            corrupt_strategy = "full"  # if random.random() < label_args.corrupt_entire_chunk_prob_full else "asr"
        else:
            corrupt_strategy = label_args.corrupt_entire_chunk_strategy

        input_text = tokenizer.decode(input_ids)
        if corrupt_strategy == "tokenizer":
            if not tokenizer:
                raise NotImplementedError()
            corrupted = corrupt(input_text, do_lowercase=True, do_remove_punct=False)
            input_ids = tokenizer.encode(corrupted, add_special_tokens=False, verbose=False)
            # remove ALL punct *tokens*
            auxiliary_remove_prob = 1.0
        elif corrupt_strategy == "full":
            # remove all punct *characters*
            corrupted = corrupt(input_text, do_lowercase=True, do_remove_punct=True)
            input_ids = tokenizer.encode(corrupted, add_special_tokens=False, verbose=False)
            auxiliary_remove_prob = 1.0  # just for safety/consistency
        elif corrupt_strategy == "asr":
            if not tokenizer:
                raise NotImplementedError()
            corrupted_sentences = corrupt_asr(input_text, lang)
            corrupted_text = "\n".join(corrupted_sentences)
            input_ids = tokenizer.encode(corrupted_text, add_special_tokens=False, verbose=False)
            auxiliary_remove_prob = 0.0  # do not remove additional characters.
        block_ids = [0] * len(input_ids)

    else:
        auxiliary_remove_prob = label_args.auxiliary_remove_prob

    labels = label(input_ids, label_dict)

    separator = Constants.SEPARATORS[lang]

    try:
        i = next(index for index, label in enumerate(labels) if label != 0)
    except StopIteration:
        return input_ids, block_ids, labels

    if tokenizer:
        # account for CLS and SEP token, added later
        min_length = min_length - 2 if min_length is not None else None
    while min_length is None or len(input_ids) > min_length:
        if labels[i] == Constants.NEWLINE_INDEX + 1:
            if random.random() < label_args.newline_remove_prob:
                if separator == " " and random.random() < label_args.newline_whitespace_prob:
                    if tokenizer:
                        # inserting " " leaks \n information
                        # the token is never there naturally, so it is a 1:1 proxy for \n
                        del input_ids[i + 1]
                        del labels[i + 1]
                        del block_ids[i + 1]
                    else:
                        input_ids[i + 1] = ord(" ")
                else:
                    del input_ids[i + 1]
                    del labels[i + 1]

                    if pack_samples:
                        last_index_in_block = i
                        while (
                            last_index_in_block + 1 == len(block_ids)
                            or last_index_in_block < len(block_ids)
                            and block_ids[last_index_in_block + 1] == block_ids[last_index_in_block]
                        ):
                            last_index_in_block += 1
                        input_ids.insert(last_index_in_block, 0)
                        labels.insert(last_index_in_block, 0)
                    else:
                        del block_ids[i + 1]
                    if (
                        tokenizer
                        and separator == ""
                        and label_args.non_whitespace_remove_spaces
                        and i + 1 < len(input_ids)
                    ):
                        # tokenizer.decode() retains the space that leaks the information
                        # so we need to get the position within the tokenized text and then remove the space
                        # (so there is no more space when fed into the tokenizer call)
                        if input_ids[i + 1] == tokenizer.convert_tokens_to_ids("▁"):
                            # remove artificial space
                            del input_ids[i + 1]
                            del labels[i + 1]
                            del block_ids[i + 1]
                        if i + 1 < len(input_ids):
                            next_token = tokenizer.convert_ids_to_tokens(input_ids[i + 1])
                            if next_token.startswith("▁"):
                                # next token starts with _ --> remove the _ from the token and re-tokenize
                                remove_next = False
                                remaining_token = tokenizer.convert_ids_to_tokens(input_ids[i + 1])
                                if len(remaining_token) > 1:
                                    # ▁Test --> Test
                                    remaining_token = remaining_token[1:]
                                else:
                                    # ▁ --> remove
                                    remove_next = True
                                remaining_id = tokenizer.convert_tokens_to_ids(remaining_token)
                                # replace the token with the remaining token
                                if remaining_id != tokenizer.unk_token_id:
                                    input_ids[i + 1] = remaining_id
                                else:
                                    # UNK token, remove it
                                    remove_next = True
                                if remove_next:
                                    del input_ids[i + 1]
                                    del labels[i + 1]
                                    del block_ids[i + 1]

        elif label_args.use_auxiliary and labels[i] > Constants.AUX_OFFSET:  # auxiliary
            if pack_samples:
                raise NotImplementedError()

            if random.random() < auxiliary_remove_prob:
                if label_args.retain_first_consecutive_punctuation:
                    # remove only if the next token is not a newline
                    # this retains the current auxiliary character, even though we decided to remove it
                    # it may skew the statistics since an auxiliary character is a better proxy for a newline
                    if labels[i + 1] != 1:
                        del input_ids[i + 1]
                        del labels[i + 1]
                        del block_ids[i + 1]
                else:
                    # in case of something like ".\n", this removes the "." and the \n label (=1)
                    # so the newline in the text is kept, but the label is removed!
                    del input_ids[i + 1]
                    del labels[i + 1]
                    del block_ids[i + 1]

        try:
            i = i + 1 + next(index for index, label in enumerate(labels[i + 1 :]) if label != 0)
        except StopIteration:
            break

    return input_ids, block_ids, labels


def indices_to_sentences(text, indices, strip_whitespace=False):
    sentences = []

    offset = 0
    idx = 0
    for idx in indices:
        idx = idx + 1
        while idx < len(text) and text[idx].isspace():
            idx += 1

        sentence = text[offset:idx]
        if strip_whitespace:
            # NB: I would have thought that this is slower than
            # adjusting the start and end indices since there are
            # two string copies, but it seems to be faster
            # (at least on short strings). more reason to port to Rust?
            sentence = sentence.strip()

        if len(sentence) > 0:
            sentences.append(sentence)

        offset = idx

    if idx != len(text):
        last_sentence = text[idx:]
        if strip_whitespace:
            last_sentence = last_sentence.strip()

        if len(last_sentence) > 0:
            sentences.append(last_sentence)

    return sentences


def reconstruct_sentences(text, partial_sentences):
    # for consistency
    partial_sentences = [x.strip() for x in partial_sentences]

    fixed_sentences = []
    i = 0
    for x in partial_sentences:
        try:
            idx = i + text[i:].index(x[:-1])
        except ValueError:
            reduced = x[:-2]

            while len(reduced) > 0 and not reduced.isspace():
                idx = i + text[i:].find(reduced)

                if idx == -1:
                    reduced = reduced[:-1]
                else:
                    break

            if idx == -1:
                raise ValueError("irrecoverable error in reconstruction")

        if idx > i:
            fixed_sentences.append(text[i:idx])
            i = idx

    if i != len(text):
        fixed_sentences.append(text[i:])

    return fixed_sentences


def get_token_spans(tokenizer, offsets_mapping, tokens):
    # Filter out special tokens and get their character start and end positions
    valid_indices = np.array(
        [
            idx
            for idx, token in enumerate(tokens)
            if token not in [tokenizer.cls_token, tokenizer.sep_token, tokenizer.pad_token]
            and idx < len(offsets_mapping)
        ]
    )
    valid_offsets = np.array(offsets_mapping)[valid_indices]
    return valid_indices, valid_offsets


def token_to_char_probs(text, tokens, token_logits, tokenizer, offsets_mapping):
    """Map from token probabalities to character probabilities"""
    char_probs = np.full((len(text), token_logits.shape[1]), -np.inf)  # Initialize with very low numbers

    valid_indices, valid_offsets = get_token_spans(tokenizer, offsets_mapping, tokens)

    # Assign the token's probability to the last character of the token
    for i in range(valid_offsets.shape[0]):
        start, end = valid_offsets[i]
        char_probs[end - 1] = token_logits[valid_indices[i]]

    return char_probs


if __name__ == "__main__":
    # test corrupt function
    from tokenizers import AddedToken

    tokenizer = AutoTokenizer.from_pretrained("xlm-roberta-base")
    tokenizer.add_special_tokens({"additional_special_tokens": [AddedToken("\n")]})
    text = "That's right, Five!\n!\n!!!\n!\n Always lay the!Blame on others!"
    input_ids = tokenizer(text)["input_ids"]
    block_ids = [0] * len(input_ids)
    label_args = LabelArgs(
        custom_punctuation_file="punctuation_xlmr_unk.txt",
        use_auxiliary=True,
        auxiliary_remove_prob=1.0,
        newline_whitespace_prob=1.0,
    )
    label_dict = get_subword_label_dict(label_args, tokenizer)
    print(len(label_dict))
    # print all tokens with a number in it (from label_dict only)
    print(
        [tokenizer.decode(input_id) for input_id in input_ids if any(c.isdigit() for c in tokenizer.decode(input_id))]
    )
    # corrupt
    input_ids, block_ids, labels = corrupt_training(
        input_ids, block_ids, "en", label_args, label_dict, tokenizer=tokenizer
    )
    print(input_ids)
    print(labels)
    print(tokenizer.tokenize(text))
    print([(tokenizer.decode([input_id]), label) for input_id, label in zip(input_ids, labels)])
